#!/usr/bin/python3
#  -*- coding: utf-8 -*-

import os
import sys
import re
import datetime
import argparse
import pprint

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import matplotlib.dates as mdates
import sklearn.cluster
import sklearn.decomposition
import scipy.stats

import datfile
import asciigrid
from utils import ValueError

bigsize = ( 10, 5 )
highsize = ( 10, 5.3 )
fatsize = ( 10, 10 )

def write_image( fig, outdir, name ) :
    if outdir is not None :
        filename = os.path.join( outdir, name )

        fig.savefig( filename, dpi=300, bbox_inches="tight" )
        print( "Written image file <{}>...".format(filename) )

    return

def draw_alldata( X, D, grid=None, outdir=None, large=False ) :
    first = D.min()

    if grid is None :
        ( xll, yll ) = ( X[:,0].min(), X[:,1].min() )
        X = X - ( xll, yll, 0 )
        
    D = (D - first) / (24. * 3600)

    start = D.min()
    end = D.max()

    if large :
        ticknum = np.linspace( start, end, 5 )
        tickepc = [ int((n * 24 * 3600) + first) for n in ticknum ]
        ticklbl = [ datetime.datetime.fromtimestamp(epc).strftime("%Y-%m") for epc in tickepc ]
    else :
        ticknum = np.linspace( start, end, 7 )
        tickepc = [ int((n * 24 * 3600) + first) for n in ticknum ]
        ticklbl = [ datetime.datetime.fromtimestamp(epc).strftime("%Y/%m/%d") for epc in tickepc ]

    title = "Cluster riconosciuti nel tempo"

    if grid is None :
        fig = plt.figure()
        ax = fig.add_subplot( 111 )
        ax.set_aspect( 'equal' )
    else :
        if large :
            fig = grid.draw( title="" )
        else :
            fig = grid.draw( title=title )
        ax = grid.axis

    sc = ax.scatter( X[:, 0], X[:, 1], s=X[:, 2], c=D, marker='o', cmap=plt.cm.jet ) #gnuplot )
    if large :
        ax.set_xlabel( "Easting (m)", size=20 )
        ax.set_ylabel( "Northing (m)", size=20 )
        ax.tick_params( axis='both', labelsize=18 )
    else :
        ax.set_xlabel( "Easting (m)" )
        ax.set_ylabel( "Northing (m)" )

    cbar = plt.colorbar( sc, location="bottom", orientation="horizontal" )
    if not large :
        cbar.set_label( "Days" )

    #cbar.set_ticks( tickepc, labels=ticklbl, size=18 )
    cbar.set_ticks( ticknum )

    if large :
        cbar.set_ticklabels( ticklbl, size=18 )
    else :
        cbar.set_ticklabels( ticklbl )

    if not large and grid is None :
        fig.suptitle( title )

    plt.draw()

    write_image( fig, outdir, "TimeClusters.png" )

    return X, D

def clusterize( X, maxdist, npts, grid=None, outdir=None, large=False ) :
    #base = ord( 'A' )

    if grid is not None :
        ( xll, yll ) = ( X[:, 0].min(), X[:, 1].min() )
        X = X - ( xll, yll, 0 )

    S = np.array( [ [x[0], x[1]] for x in X ] )
    lbs = sklearn.cluster.DBSCAN( eps=maxdist, min_samples=npts ).fit_predict( S )

    labels = lbs.flatten()
    unique_labels = set( labels )
    clusnum = len( unique_labels )
    if -1 in labels :
        clusnum -= 1
        noisenum = list( labels ).count( -1 )
    else :
        noisenum = 0

    print( "Found {} clusters and {} noise...".format(clusnum, noisenum) )
    title = "Zone aggregate ({})".format(clusnum)

    lbls = np.array( [ l for l in unique_labels if l != -1 ] )
    colormap = plt.cm.gist_ncar #tab20
    norm = colors.Normalize( vmin=lbls.min(), vmax=lbls.max() )
    mapper = plt.cm.ScalarMappable( norm=norm, cmap=colormap )

    if grid is None :
        fig = plt.figure()
        ax = fig.add_subplot( 111 )
        ax.set_aspect( 'equal' )
    else :
        if large :
            fig = grid.draw( title="" )
        else :
            fig = grid.draw( title=title )
        ax = grid.axis

        X = X + ( xll, yll, 0 )

    mask = (labels != -1)
    N = X[labels == -1]
    L = labels[mask]
    C = X[mask]

    zones = []

    ax.scatter( N[:, 0], N[:, 1], N[:, 2], c='black', marker='o' )
    #sc = ax.scatter( C[:, 0], C[:, 1], C[:, 2], c=L, marker='o', cmap=colormap )

    for lbl in unique_labels :
        if lbl == -1 :
            continue

        name = "Z{:03d}".format( lbl ) #chr( base + lbl )
        Z = X[labels==lbl]

        n = len( Z )
        x = Z[:, 0].sum() / float(n)
        y = Z[:, 1].sum() / float(n)
        ax.scatter( Z[:, 0], Z[:, 1], Z[:, 2], marker='o', color=mapper.to_rgba(lbl), edgecolors='black', linewidths=0.5, label="Zone {}".format(name) )
        ax.text( x, y, name, color='yellow', fontsize='large', weight='bold' ) 

        zones.append( dict( bary=(x, y) ) )

    if large :
        ax.set_xlabel( "Easting (m)", size=20 )
        ax.set_ylabel( "Northing (m)", size=20 )
        ax.tick_params( axis='both', labelsize=18 )
    else :
        ax.set_xlabel( "Easting (m)" )
        ax.set_ylabel( "Northing (m)" )
    #ax.legend( fontsize='medium', loc="upper right", bbox_to_anchor=(1.1, 0.99) )

    if not large and grid is None :
        fig.suptitle( title )

    plt.draw()

    write_image( fig, outdir, "ZonedClusters.png" )

    return labels, zones

def show_zones( labels, clusters, summary, outdir=None ) :
    #base = ord( 'A' )
    unique = set( labels )
    clusters = np.array( clusters )

    if outdir is None :
        handle = sys.stdout
        csvlines = None
    else :
        filename = os.path.join( outdir, "SummaryZones.dat" )
        handle = open( filename, "w" )
        csvlines = []

    print( "#Zona\tID\tUTM_x (m)\tUTM_y (m)\tN\tPmean (W)\tFmean (W/m²)", file=handle )

    for lbl in unique :
        if lbl == -1 :
            continue

        if csvlines is not None :
            if len(csvlines) != 0 :
                csvlines.append( [] )

            csvlines.append( [ "Zona", "ID", "UTM_x (m)", "UTM_y (m)", "NC", "Tm (°C)", "dTm (°C)", "Pm (W)", "dPm (W)", "Fm (W/m²)", "dFm (W/m²)" ] )

        zc = clusters[labels == lbl]
        zc = [ c for c in zc ]
        zc.sort( key=lambda c: c.date )

        power = ValueError.sum_l2( [ c.power for c in zc ] )
        tmean = ValueError.average( [ c.tmean for c in zc ] )
        #flux  = ValueError.average( [ c.flux.value for c in zc ] )
        flux  = ValueError.average( [ c.flux for c in zc ] )

        summary[lbl]['power'] = power.copy()
        summary[lbl]['flux'] = flux.copy()
        summary[lbl]['tmean'] = tmean.copy()

        name = "Z{:03d}".format(lbl)
        baseline = [ name, str(lbl), "{:.2f}".format(summary[lbl]['bary'][0]), "{:.2f}".format(summary[lbl]['bary'][1]), str(len(zc)) ]
        if csvlines is not None :
            csvlines.append( baseline + [ str(tmean.value), str(tmean.error), str(power.value), str(power.error), str(flux.value), str(flux.error) ] )
        line = baseline + [ str(power), str(flux) ]
        
        print( "\t".join(line), file=handle )

        if csvlines is not None :
            csvlines.append( [ "Clusters" ] )
            csvlines.append( [ "", "ID", "Nome", "Prodotto", "Data", "UTM_x (m)", "UTM_y (m)", "Lon", "Lat", "R (m)", "dR (m)", "Area m²", "Tm (°C)", "dTm (°C)", "P (W)", "dP (W)", "F (W/m²)", "dF (W/m²)" ] )
            for c in zc :
                line = [ "", str(c.cid), c.name, str(int(c.prod)), c.date.strftime("%Y/%m/%d"),
                         "{:.2f}".format(c.utm[0]), "{:.2f}".format(c.utm[1]),
                         "{:.7f}".format(c.lonlat[0]), "{:.7f}".format(c.lonlat[1]),
                         str(c.rmet.value), str(c.rmet.error), "{:.1f}".format(c.marea),
                         str(c.tmean.value), str(c.tmean.error), str(c.power.value), str(c.power.error),
                         str(c.flux.value), str(c.flux.error) ]

                csvlines.append( line )

    if outdir is not None :
        handle.close()
        print( "Written file <{}>".format(filename) )

        if len(csvlines) != 0 :
            filename = os.path.join( outdir, "ZoneInfo.csv" )
            with open( filename, "w" ) as handle :
                for l in csvlines :
                    line = ";".join(l).replace( ".", "," )
                    print( line, file=handle )

            print( "Written file <{}>".format(filename) )

    return summary

def add_rsam( fig, ax, RSAM ) :
    ax3 = fig.add_axes( ax.get_position(), frameon=False )
    ax3.set_zorder( -5 )
    ax.patch.set_visible( False )
    ax3.set_xlim( ax.get_xlim() )
    ax3.set_ylim( 0, 40000 )
    ax3.tick_params( axis='both', labeltop=False, labelbottom=False, labelleft=False, labelright=False,
                     top=False, bottom=False, left=False, right=False )

    ax3.plot( RSAM[:, 0], RSAM[:, 1], '.g', ms=0.3 )

    return

def simple_evolution( data, clusters, labels, summary, outdir=None, RSAM=None, large=False ) :
    unique = set( labels )

    lbls = np.array( [ l for l in unique if l != -1 ] )

    norm = colors.Normalize( vmin=lbls.min(), vmax=lbls.max() )
    mapper = plt.cm.ScalarMappable( norm=norm, cmap=plt.cm.gist_ncar ) #tab20 )
    
    bgs = { d['date']: ( d['Tbg'], d['dTbg'], d['Tair'], d['dTair'] ) for d in data }
    arrays = dict()
    zcols = dict()

    C = np.array( clusters )

    for lbl in unique :
        if lbl == -1 :
            continue

        zname = "Z{:03d}".format(lbl)
        zinfo = dict()

        zcols[zname] = mapper.to_rgba( lbl )

        zc = C[labels == lbl]
        for c in zc :
            if c.date in zinfo :
                zinfo[c.date].append( c )
            else :
                zinfo[c.date] = [ c ]

        times   = []
        sizes   = []
        tmeans  = []
        dtmeans = []
        powrs   = []
        dpwrs   = []
        tbgs    = []
        dtbgs   = []
        tair    = []
        dtair   = []

        for date, cs in sorted(zinfo.items()) :
            times.append( date )
            sdate = date.strftime( "%Y/%m/%d" )

            size = np.array( [ c.marea for c in cs ] ).sum()
            power = ValueError.sum_l2( [ c.power for c in cs ] ) / size
            #power = ValueError.average( [ c.power for c in cs ] )
            tmean = ValueError.average( [ c.tmean for c in cs ] )
        
            sizes.append( size )
            powrs.append( power.value )
            dpwrs.append( power.error )
            tmeans.append( tmean.value )
            dtmeans.append( tmean.error )
            tbgs.append( bgs[sdate][0] )
            dtbgs.append( bgs[sdate][1] )
            tair.append( bgs[sdate][2] )
            dtair.append( bgs[sdate][3] )

        arrays[zname] = np.array( [ times, sizes, powrs, dpwrs, tbgs, dtbgs, tair, dtair, tmeans, dtmeans ] )

    figure = plt.figure( figsize=bigsize )
    ax0 = figure.add_subplot( 211 )
    ax1 = figure.add_subplot( 212 )

    for zone in arrays :
        ax0.plot( arrays[zone][0, :], arrays[zone][1, :], marker='o', label="Zone {}".format(zone), lw=1, color=zcols[zone] )
        ax1.errorbar( arrays[zone][0, :], arrays[zone][2, :], arrays[zone][3, :], label="Zone {}".format(zone), lw=1, color=zcols[zone] )

    ax0.set_ylabel( "Area (m$^2$)" )
    ax1.set_ylabel( "$\Phi$ (W/m$^2$)" )

    ax0.grid()
    ax1.grid()

    if not large :
        figure.suptitle( "Zone trend" )

    write_image( figure, outdir, "AllZones.png" )

    total_sizes = dict()

    if outdir is not None :
        fig = plt.figure( figsize=bigsize )
        ax1 = fig.add_subplot( 111 )

        for zone in arrays :
            ax1.errorbar( arrays[zone][6, :], arrays[zone][8, :], arrays[zone][9, :], arrays[zone][7, :], lw=0.5, ls='', marker='o', color=zcols[zone] )

        ax1.set_xlabel( "Air temperature (°C)" )
        ax1.set_ylabel( "Air temperatura (°C)" )
        ax1.grid()

        if not large :
            fig.suptitle( "$T_m$ vs $T_a$ (°C)\nAll zones" )

        write_image( fig, outdir, "MediaVsAria.png" )

        plt.close( fig )

        decorr = dict()

        for zone in arrays :
            times = arrays[zone][0, :].astype( datetime.datetime )
            sizes = arrays[zone][1, :].astype( float )
            for (t, s) in zip(times, sizes) :
                if t in total_sizes :
                    total_sizes[t] += s
                else :
                    total_sizes[t] = s

            fig = plt.figure( figsize=bigsize )
            ax1 = fig.add_subplot( 111 )

            coeff = np.corrcoef( arrays[zone][2, :].astype(float), arrays[zone][4, :].astype(float) )[0, 1]

            tmeans = arrays[zone][8, :].astype( float )
            tair   = arrays[zone][6, :].astype( float )

            if len(tmeans) <= 2 :
                print( "Zone {}: too few points ({}) to correlate data...".format(zone, len(tmeans)) )
                continue
            else :
                tcoff = np.corrcoef( tair, tmeans )[0, 1]
                corr, pval = scipy.stats.pearsonr( tair, tmeans )

                print( "Zone {}: flux correlation: {:.2f}, T correlation: {:.2f} ({:.2f}, {:.2f})".format(zone, coeff, tcoff, corr, pval) )

            X = np.array( [ tair, tmeans ] ).transpose()
            model = sklearn.decomposition.PCA()
            features = model.fit_transform( X )

            dtair = features[:, 0]
            dtmean = features[:, 1]
            decorr[zone] = dtmean

            ax1.errorbar( arrays[zone][0, :], arrays[zone][2, :], arrays[zone][3, :], marker='o', lw=0.5, color=zcols[zone] )
            ax1.tick_params( axis='y', labelcolor=zcols[zone] )
            if large :
                ax1.set_ylabel( "$\Phi$ (W/m$^2$)", color=zcols[zone], size=20 )
                ax1.tick_params( axis='both', labelsize=18 )
                ax1.xaxis.set_major_locator( plt.MaxNLocator(5) )
                ax1.xaxis.set_major_formatter( mdates.DateFormatter("%Y-%m") )
            else :
                ax1.set_ylabel( "$\Phi$ (W/m$^2$)", color=zcols[zone] )

            ax1.grid()

            ax2 = ax1.twinx()
            ax2.errorbar( arrays[zone][0, :], arrays[zone][4, :], arrays[zone][5, :], marker='+', lw=0.5, color='red', label="T$_{bg}$" )
            ax2.errorbar( arrays[zone][0, :], arrays[zone][6, :], arrays[zone][7, :], marker='x', lw=0.5, color='red', label="T$_{air}$" )
            ax2.tick_params( axis='y', labelcolor='red' )
            if large :
                ax2.set_ylabel( "T ($^o$C)", color='red', size=20 )
                ax2.legend( fontsize=18 )
            else :
                ax2.set_ylabel( "T ($^o$C)", color='red' )
                ax2.legend()

            if RSAM is not None :
                add_rsam( fig, ax1, RSAM )

            if not large :
                fig.suptitle( "Heat flux zone {} (C={:.2f}, $\Phi_{{mean}}$={} W/m$^2$)".format(zone, coeff, summary[lbl]['flux']) )

            write_image( fig, outdir, "FlussoZona_{}.png".format(zone) )

            plt.close( fig )

            fig = plt.figure( figsize=bigsize )
            ax1 = fig.add_subplot( 111 )

            ax1.plot( arrays[zone][0, :], arrays[zone][1, :], marker='o', lw=0.5, color=zcols[zone] )
            ax1.set_ylabel( "Area (m$^2$)" )
            ax1.grid()

            ax2 = ax1.twinx()
            ax2.errorbar( arrays[zone][0, :], arrays[zone][4, :], arrays[zone][5, :], marker='+', lw=0.5, color='red', label="T$_{bg}$" )
            ax2.errorbar( arrays[zone][0, :], arrays[zone][6, :], arrays[zone][7, :], marker='x', lw=0.5, color='red', label="T$_{air}$" )
            ax2.set_ylabel( "Temperature ($^o$C)", color='red' )
            ax2.tick_params( axis='y', labelcolor='red' )
            ax2.legend()

            if RSAM is not None :
                add_rsam( fig, ax1, RSAM )

            if not large :
                fig.suptitle( "Area zone {}".format(zone) )

            write_image( fig, outdir, "AreaZona_{}.png".format(zone) )

            plt.close( fig )

            fig = plt.figure( figsize=fatsize )
            ax1 = fig.add_subplot( 211 )

            ax1.errorbar( arrays[zone][0, :], arrays[zone][8, :], arrays[zone][9, :], marker='o', lw=0.5, color=zcols[zone] )
            if large :
                ax1.set_ylabel( "T$_{avg}$ (°C)", size=20 )
                ax1.tick_params( axis='both', labelsize=18 )
                ax1.xaxis.set_major_locator( plt.MaxNLocator(5) )
                ax1.xaxis.set_major_formatter( mdates.DateFormatter("%Y-%m") )
                #ax1.tick_params( axis='x', labelrotation=60 )
            else :
                ax1.set_ylabel( "T$_{avg}$ (°C)" )
            ax1.grid()

            ax2 = ax1.twinx()
            ax2.errorbar( arrays[zone][0, :], arrays[zone][4, :], arrays[zone][5, :], marker='+', lw=0.5, color='red', label="T$_{bg}$" )
            ax2.errorbar( arrays[zone][0, :], arrays[zone][6, :], arrays[zone][7, :], marker='x', lw=0.5, color='red', label="T$_{air}$" )
            ax2.tick_params( axis='y', labelcolor='red' )
            if large :
                ax2.set_ylabel( "T ($^o$C)", color='red', size=20 )
                ax2.tick_params( axis='both', labelsize=18 )
                ax2.legend( fontsize=18 )
            else :
                ax2.set_ylabel( "T ($^o$C)", color='red' )
                ax2.legend()

            ax1 = fig.add_subplot( 212 )

            ax1.errorbar( arrays[zone][0, :], dtmean, arrays[zone][9, :], marker='o', lw=0.5, color=zcols[zone] )
            if large :
                ax1.set_ylabel( "T$_{avg}$ residual (°C)", size=20 )
                ax1.tick_params( axis='both', labelsize=18 )
                ax1.xaxis.set_major_locator( plt.MaxNLocator(5) )
                ax1.xaxis.set_major_formatter( mdates.DateFormatter("%Y-%m") )
                #ax1.tick_params( axis='x', labelrotation=60 )
            else :
                ax1.set_ylabel( "T$_{avg}$ residual (°C)" )
            ax1.grid()

            if RSAM is not None :
                add_rsam( fig, ax1, RSAM )

            if not large :
                fig.suptitle( "Temperature zone {}".format(zone) )

            write_image( fig, outdir, "TemperaturaZona_{}".format(zone) )

            plt.close( fig )

            fig = plt.figure( figsize=fatsize )
            ax1 = fig.add_subplot( 211 )

            ax1.errorbar( arrays[zone][6, :], arrays[zone][8, :], arrays[zone][9, :], arrays[zone][7, :], lw=0.5, ls='', marker='o', color=zcols[zone] )
            ax1.set_xlabel( "Air temperature (°C)" )
            ax1.set_ylabel( "Average temperatura (°C)" )
            ax1.grid()

            ax2 = fig.add_subplot( 212 )

            ax2.errorbar( dtair, dtmean, arrays[zone][9, :], arrays[zone][7, :], lw=0.5, ls='', marker='o', color=zcols[zone] )
            ax2.set_xlabel( "Decorrelated T$_a$ (°C)" )
            ax2.set_ylabel( "Decorrelated T$_m$ (°C)" )
            ax2.grid()

            if not large :
                fig.suptitle( "Average temperature vs Air temperature - zone {}\nCorrelation {:.2f}".format(zone, tcoff) )

            write_image( fig, outdir, "MediaVsAriaZona_{}.png".format(zone) )

            plt.close( fig )

        filename = os.path.join( outdir, "decorrelated.dat" )
        with open( filename, "w" ) as handle :
            pprint.pprint( decorr, stream=handle )

        fig = plt.figure( figsize=highsize )
        ax1 = fig.add_subplot( 111 )

        ncols = 0
        for zone in decorr :
            if len(decorr[zone]) >= 25 :
                ncols += 1
                ax1.errorbar( arrays[zone][0, :], decorr[zone], arrays[zone][9, :], marker='o', lw=0, color=zcols[zone], label=zone )

        if large :
            ax1.set_ylabel( "T$_{avg}$ residual ($^o$C)", size=20 )
            ax1.tick_params( axis='both', labelsize=18 )
            ax1.xaxis.set_major_locator( plt.MaxNLocator(5) )
            ax1.xaxis.set_major_formatter( mdates.DateFormatter("%Y-%m") )
            ax1.legend( fontsize=13, ncol=ncols, loc="lower center", bbox_to_anchor=(0.5, 1.1) )
        else :
            ax1.set_ylabel( "T$_{avg}$ residual ($^o$C)" )
            ax1.legend()
        ax1.grid()

        if not large :
            fig.suptitle( "Decorrelated temperatures" )

        write_image( fig, outdir, "TemperatureDecorrelate" )

        plt.close( fig )                

    return [ (t, total_sizes[t]) for t in total_sizes ]

def total_evolution( data, outdir=None, RSAM=None, large=False ) :
    A = np.array( [ [ datetime.datetime.fromtimestamp(d['epoch']), d['Fred'], d['dFred'], d['Tbg'], d['dTbg'], d['Ftot'], d['dFtot'], d['Tair'], d['dTair'], d['Tmax'], d['dTmax'] ] for d in data ] )

    flux = ValueError.average( A[:, 1] )
    coeff = np.corrcoef( A[:, 1].astype(float), A[:, 3].astype(float) )

    fig = plt.figure( figsize=bigsize )
    ax1 = fig.add_subplot( 111 )
    
    ax1.errorbar( A[:, 0], A[:, 1], A[:, 2], marker='o', color='blue', lw=0.5 )
    ax1.set_ylabel( "$\Phi$ (W/m$^2$)", color='blue' )
    ax1.tick_params( axis='y', labelcolor='blue' )
    ax1.grid()

    ax2 = ax1.twinx()
    ax2.errorbar( A[:, 0], A[:, 3], A[:, 4], marker='+', color='red', lw=0.5, label="T$_{bg}$" )
    ax2.errorbar( A[:, 0], A[:, 7], A[:, 8], marker='x', color='red', lw=0.5, label="T$_{air}$" )
    ax2.set_ylabel( "T ($^o$C)", color='red' )
    ax2.tick_params( axis='y', labelcolor='red' )
    ax2.legend()

    if RSAM is not None :
        add_rsam( fig, ax1, RSAM )

    if not large :
        fig.suptitle( "Reduced flux: global trend (C={:.2f}, $\Phi_{{mean}}$={} W/$^2$)".format(coeff[0,1], flux) )

    write_image( fig, outdir, "FlussoGlobaleRidotto.png" )

    fig = plt.figure()

    plt.errorbar( A[:, 3], A[:, 1], A[:, 2], A[:, 4], marker='+', ms=4, lw=1, ls='none', label="T$_{bg}$" )
    plt.errorbar( A[:, 7], A[:, 1], A[:, 2], A[:, 8], marker='x', ms=4, lw=1, ls='none', label="T$_{air}$" )
    plt.xlabel( "T ($^o$C)" )
    plt.ylabel( "$\Phi_{red}$ (W$^2$)" )
    plt.legend()

    if not large :
        fig.suptitle( "Flux vs. T$_bg$ (All flights, C={:.2f})".format(coeff[0,1]) )
    plt.grid()
    
    write_image( fig, outdir, "FlussoRidotto_vs_Tbg.png" )

    flux = ValueError.average( A[:, 5] )
    coeff = np.corrcoef( A[:, 5].astype(float), A[:, 3].astype(float) )

    fig = plt.figure( figsize=bigsize )
    ax1 = fig.add_subplot( 111 )

    ax1.errorbar( A[:, 0], A[:, 5], A[:, 6], marker='o', color='blue', lw=0.5 )
    ax1.set_ylabel( "$\Phi$ (W/m$^2$)", color='blue' )
    ax1.tick_params( axis='y', labelcolor='blue' )
    ax1.grid()

    ax2 = ax1.twinx()
    ax2.errorbar( A[:, 0], A[:, 3], A[:, 4], marker='+', color='red', lw=0.5, label="T$_{bg}$" )
    ax2.errorbar( A[:, 0], A[:, 7], A[:, 8], marker='x', color='red', lw=0.5, label="T$_{air}$" )
    ax2.set_ylabel( "T ($^o$C)", color='red' )
    ax2.tick_params( axis='y', labelcolor='red' )
    ax2.legend()

    if RSAM is not None :
        add_rsam( fig, ax1, RSAM )

    if not large :
        fig.suptitle( "Total flux: global trend (C={:.2f}, $\Phi_{{mean}}$={} W/$^2$)".format(coeff[0,1], flux) )

    write_image( fig, outdir, "FlussoGlobaleTotale.png" )

    fig = plt.figure( figsize=bigsize )
    ax1 = fig.add_subplot( 111 )

    ax1.errorbar( A[:, 0], A[:, 9], A[:, 10], marker='o', color='red', lw=0.5 )
    ax1.set_ylabel( "T$_{max}$ ($^o$C)", color='red' )
    ax1.tick_params( axis='y', labelcolor='red' )
    ax1.grid()

    ax2 = ax1.twinx()
    ax2.errorbar( A[:, 0], A[:, 3], A[:, 4], marker='+', color='blue', lw=0.5, label="T$_{bg}$" )
    ax2.errorbar( A[:, 0], A[:, 7], A[:, 8], marker='x', color='blue', lw=0.5, label="T$_{air}$" )
    ax2.set_ylabel( "T ($^o$C)", color='blue' )
    ax2.tick_params( axis='y', labelcolor='blue' )
    ax2.legend()

    if RSAM is not None :
        add_rsam( fig, ax1, RSAM )

    if not large :
        fig.suptitle( "Maximum temperature" )

    write_image( fig, outdir, "TemperaturaMassima.png" )

    return

def total_area( data, sizes, outdir=None ) :
    A = np.array( [ [ datetime.datetime.fromtimestamp(d['epoch']), d['Tair'], d['dTair'], d['Tbg'], d['dTbg'] ] for d in data ] )
    S = np.array( sizes )

    fig = plt.figure( figsize=bigsize )
    ax1 = fig.add_subplot( 111 )

    ax1.plot( S[:, 0], S[:, 1], marker='o', color='red', lw=0.5, ls='-' )
    ax1.set_ylabel( "Total area (m$^2$)" )
    ax1.grid()

    ax2 = ax1.twinx()
    ax2.errorbar( A[:, 0], A[:, 1], A[:, 2], marker='+', color="blue", lw=0.5, label="T$_{air}$" )
    ax2.errorbar( A[:, 0], A[:, 3], A[:, 4], marker='+', color="blue", lw=0.5, label="T$_{air}$" )
    ax2.set_ylabel( "T ($^o$C)", color='blue' )
    ax2.tick_params( axis='y', labelcolor='blue' )
    ax2.legend()

    fig.suptitle( "Area totale" )

    write_image( fig, outdir, "AreaTotale.png" )
    
    return

def dummy() :
    fig = plt.figure( figsize=bigsize )
    ax1 = fig.add_subplot( 111 )

    ax1.plot( A[:, 0], A[:, 11], marker='o', color='red', lw=0.5, ls='-' )
    ax1.set_ylabel( "Total area (m$^2$)" )
    ax1.grid()

    ax2 = ax1.twinx()
    ax2.errorbar( A[:, 0], A[:, 3], A[:, 4], marker='+', color='blue', lw=0.5, label="T$_{bg}$" )
    ax2.errorbar( A[:, 0], A[:, 7], A[:, 8], marker='x', color='blue', lw=0.5, label="T$_{air}$" )
    ax2.set_ylabel( "T ($^o$C)", color='blue' )
    ax2.tick_params( axis='y', labelcolor='blue' )
    ax2.legend()

    if RSAM is not None :
        add_rsam( fig, ax1, RSAM )

    fig.suptitle( "Area totale" )

    write_image( fig, outdir, "AreaTotale.png" )

    return

def read_rsam( filenames ) :
    rsam = []

    for name in filenames :
        print( "Reading file <{}>".format(name) )
        with open( name, "r" ) as handle :
            for line in handle :
                if line.startswith('#') :
                    continue

                part = line.split()

                try :
                    current = datetime.datetime.strptime( part[0], "%d/%m/%YT%H:%M:%S" )
                except :
                    current = datetime.datetime.strptime( part[0], "%Y/%m/%dT%H:%M:%S" )
                    
                vertical = float( part[1] )
                rsam.append( (current, vertical) )

    rsam.sort()

    return np.array( rsam )

def analyze( files, maxdist=20, npts=3, bgimage=None, outdir=None, rsamfiles=None, noplot=False, bbox=None, large=False ) :
    data = []
    grid = None

    #if large :
    #    size = 20
    #    print( "Setting matplotlib font size to {}".format(size) )
    #    plt.rc( 'font', size=size )

    if (rsamfiles is not None) and (len(rsamfiles) != 0) :
        RSAM = read_rsam( rsamfiles )
    else :
        RSAM = None

    if bgimage is not None :
        grid = asciigrid.AsciiGrid( bgimage, True )
        if grid.good :
            grid.colormap = plt.cm.gist_gray
            grid.usebar = False
        else :
            grid = None
    
    for filename in files :
        datafile = datfile.DatFile( filename )
        fd = datafile.read( bbox )

        if fd is not None :
            if 'clusters' in fd : 
                date = datetime.datetime.strptime( fd['date'], "%Y/%m/%d" )
                fd['epoch'] = date.timestamp()
            
                for c in fd['clusters'] :
                    c.set_date( date, fd['epoch'] )

                print( "File <{}> contains {} clusters...".format(filename, len(fd['clusters'])) )
                data.append( fd )
            else :
                print( "No cluster loaded from file <{}>...".format(filename) )
        else :
            print( "Error loading file <{}>...".format(filename) )

    data.sort( key = lambda d: d['epoch'] )

    X = []
    D = []
    R = []

    clusters = []

    for d in data : 
        X.extend( [ ( c.utm[0], c.utm[1], 0.25 * c.rmet.value * c.rmet.value ) for c in d['clusters'] ] )
        D.extend( [ float(d['epoch']) ] * len(d['clusters']) )

        clusters.extend( d['clusters'] )

    X = np.array( X )
    D = np.array( D )

    X, D = draw_alldata( X, D, grid, outdir, large=large )
    labels, zones = clusterize( X, maxdist, npts, grid, outdir, large=large )
    zones = show_zones( labels, clusters, zones, outdir )

    total_evolution( data, outdir, RSAM, large=large )
    sizes = simple_evolution( data, clusters, labels, zones, outdir, RSAM, large=large )
    total_area( data, sizes, outdir )

    if not noplot :
        plt.show()
        
    return

if __name__ == "__main__" :
    maxdist = 20
    npts    = 3
    
    parser = argparse.ArgumentParser( description="Programma di analisi dei risultati della clusterizzazione" )

    parser.add_argument( '-b', "--background", help="Carica un file ASCII GRID come background delle mappe", metavar="BGIMAGE" )
    parser.add_argument( '-m', "--maxdist", help="Distanza massima, in metri, tra due cluster per considerarli parte della medesima zona. Il default è {:.1f}m".format(maxdist),
                         metavar="MAXDIST", type=float, default=maxdist )
    parser.add_argument( '-n', "--npts", help="Numero minimo di cluster che compongono una zona. Default è {}".format(npts),
                         metavar="NPOINTS", type=int, default=npts )
    parser.add_argument( '-o', "--output", help="Directory dove vengono scritte le immagini. Se non specificato non vengono scritte",
                         metavar="OUTDIR", type=str )
    parser.add_argument( '-R', "--rsam", help="Passa uno o più file contenenti l'RSAM del dato sismico",
                         metavar="RSAMFILE", type=str, nargs='+' )
    parser.add_argument( '-i', "--files", help="Insieme di file di output della clusterizzazione da utilizzare per l'analisi", metavar="OUTDATA", nargs='+' )
    parser.add_argument( '-N', "--noplot", help="Apre le finestre interattive dei plot", action='store_true' )
    parser.add_argument( '-L', "--large", help="Draws larger axis labels", action='store_true' )
    parser.add_argument( "--bbox", help="Coordinate di esclusione dei cluster da leggere", metavar="BBOX", type=float, nargs=4 )

    arguments = parser.parse_args()

    params = dict( files=arguments.files, bgimage=arguments.background, outdir=arguments.output, rsamfiles=arguments.rsam,
                   maxdist=arguments.maxdist, npts=arguments.npts,
                   noplot=arguments.noplot, bbox=arguments.bbox, large=arguments.large )

    print( "Command line:\n{}".format(" ".join(sys.argv)) )
    analyze( **params )
