#!/usr/bin/python3
#  -*- coding: utf-8 -*-

import os
import re
import math

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as tckr

import logger
from utils import ValueError, Pixel

class AsciiGrid :
    # Class member
    rescale = 100.
    
    verbose = False
    large = False
    figsize = (9, 5)
    
    good = usenan = False
    usebar = True
    filename = None
    error = "Generic"

    columns = rows = None
    xllcorner = yllcorner = cellsize = None
    nodata = minimum = maximum = float( 'NaN' )
    countgood = countnodata = 0
    ndmask = shadow = None
    avg = None

    colormap = plt.cm.gnuplot2
    
    def __init__( self, filename, usenan=False ) :
        if os.path.exists(filename) :
            self._read_file( filename, usenan )
        else :
            error = "File <{}> does not exist".format( filename )

        return

    @classmethod
    def set_large( cls, l=True ) :
        AsciiGrid.large = l

        if l :
            AsciiGrid.figsize = (10, 10)
        else :
            AsciiGrid.figsize = (9, 5)

        return

    @classmethod
    def set_verbose( cls, v=True ) :
        AsciiGrid.verbose = v

        return

    @classmethod
    def use_figsize( cls, size ) :
        AsciiGrid.figsize = size

        return

    def _optimize_figsize( self ) :
        if self.figsize[0] < self.figsize[1] :
            ratio = float(self.rows) / float(self.columns)

            width = self.figsize[0]
            height = ratio * width

            if height < self.figsize[1] :
                height = self.figsize[1]
                width = height / ratio
        else :
            ratio = float(self.columns) / float(self.rows)

            height = self.figsize[1]
            width = ratio * height

            if width < self.figsize[0] :
                width = self.figsize[0]
                height = width / ratio

        self.figsize = ( width, height )

        return

    def cut_under( self, threshold ) :
        newval = self.minimum - AsciiGrid.rescale

        self.shadow = np.array( self.data )

        if self.usenan :
            image = np.array( self.data )
            image[np.isnan(image)] = newval
            image[image < threshold] = newval

            image[image < self.minimum] = float( 'NaN' )
            self.data = image
        else :
            self.data[self.data < threshold] = newval

        return self

    def _read_file( self, filename, usenan ) :
        self.figsize = AsciiGrid.figsize
        self.filename = filename

        firstrow = matrix = None
        head = True
        head_expr = re.compile( "^([a-zA-Z_]+)\s+(-?[0-9,\.]+)\s*$" )
        data_expr = re.compile( "^[0-9\s,\.\-eE]+$" )
        nodata_expr = re.compile( "[^0-9\s,\.\-]" )
        linen = rows = columns = 0

        with open( filename, "r" ) as handle :
            self.good = True
            
            for line in handle : # .readlines() :
                linen += 1

                if head :
                    result = re.match( head_expr, line )

                    if result :
                        param = result.group( 1 )

                        if param == "ncols" :
                            self.columns = int( result.group(2) )
                        elif param == "nrows" :
                            self.rows = int( result.group(2) )
                        elif param == "xllcorner" :
                            value = result.group(2).replace(",", ".").strip()
                            self.xllcorner = float( value )
                        elif param == "yllcorner" :
                            value = result.group(2).replace(",", ".").strip()
                            self.yllcorner = float( value )
                        elif param == "cellsize" :
                            value = result.group(2).replace(",", ".").strip()
                            self.cellsize = float( value )
                        elif param == "NODATA_value" :
                            value = result.group(2).replace(",", ".").strip()
                            self.nodata = float( value )
                        else :
                            self.error = "Illegal header keyword \"{}\" found at line {}".format(param, linen)
                            self.good = False

                            break
                    elif re.match(data_expr, line) :
                        firstrow = [ float(x) for x in line.replace(",", ".").split() ]
                        columns = len( firstrow )
                        rows = 1
                        
                        head = False
                    else :
                        self.error = "Syntax error while reading header at line {} at <<{}>>".format(linen, line.rstrip())
                        self.good = False
                        
                        break
                elif firstrow is not None :
                    matrix = firstrow
                    firstrow = None

                    row = [ float(x) for x in line.replace(",", ".").split() ]
                    if len(row) != columns :
                        self.error = "Wrong column number {} at line {} (row {}): previously found {}, all rows must have the same number of columns!".format(
                            len(row), linen, (rows + 1), columns)
                        self.good = False
                    else :
                        matrix.extend( row )

                        rows += 1
                elif re.match(data_expr, line) :
                    row = [ float(x) for x in line.replace(",", ".").split() ]
                    if len(row) != columns :
                        self.error = "Wrong column number {} at line {} (row {}): previously found {}, all rows must have the same number of columns!".format(
                            len(row), linen, (rows + 1), columns)
                        self.good = False
                    else :
                        matrix.extend( row )

                        rows += 1
                else :
                    search = re.search( nodata_expr, line )
                    self.error = "Syntax error at line {} at <<{}>>\n{}".format(linen, line.rstrip(), search)
                    self.good = False
                    
                    break

        if self.good :
            self.usenan = usenan
            
            if (rows == self.rows) and (columns == self.columns) :
                if len(matrix) == (rows * columns) :
                    mmin = mmax = None
                    
                    for i in range(len(matrix)) :
                        if matrix[i] == self.nodata :
                            if usenan :
                                matrix[i] = float( 'NaN' )

                            self.countnodata += 1
                        elif (mmin is None) or (mmax is None) :
                            mmin = mmax = matrix[i]

                            self.countgood += 1
                        else :
                            if mmin > matrix[i] :
                                mmin = matrix[i]

                            if mmax < matrix[i] :
                                mmax = matrix[i]

                            self.countgood += 1

                    self.data = np.array( matrix ).reshape( (rows, columns) )
                    if usenan :
                        mask = np.isnan( self.data )
                    else :
                        mask = self.data == self.nodata
                        self.data[mask] = mmin - AsciiGrid.rescale

                    self.ndmask = mask.reshape( (rows, columns) )
                    self.minimum = mmin
                    self.maximum = mmax

                    if (self.xllcorner == 0) or (self.yllcorner == 0) or (self.cellsize == 0) :
                        self.extent = None
                    else :
                        self.extent = [ self.xllcorner, self.xllcorner + (self.columns * self.cellsize),
                                        self.yllcorner, self.yllcorner + (self.rows * self.cellsize) ]

                    self._optimize_figsize()
                else :
                    self.error = "Wrong number of data read: {}, should be: {}".format(len(matrix), (rows * columns))
                    self.good = False
            else :
                self.error = "Wrong number of data rows or columns read. Header geometry was ({} x {}), data geometry was ({} x {})".format(self.columns, self.rows, columns, rows)
                self.good = False

        return

    def update_image( self, data, minval, maxval ) :
        if self.good :
            self.data = np.array( data ).reshape( (self.rows, self.columns) )
            self.minimum = minval
            self.maximum = maxval

            if self.usenan :
                self.data[self.ndmask] = float( 'NaN' )
            else :
                self.data[self.ndmask] = self.minimum - AsciiGrid.rescale

        return self

    def is_good( self ) :
        return self.good

    def last_error( self ) :
        return self.error

    def mean( self ) :
        m = float( 'NaN' )
        if not np.isnan(self.minimum) and not np.isnan(self.maximum) :
            m = (self.minimum + self.maximum) / 2

        return m

    def average( self, mask=None, force=False ) :
        if force or (self.avg is None) :
            if mask is None :
                data = self.data[~self.ndmask].reshape( -1, 1 )
            else :
                data = self.data[mask & ~self.ndmask].reshape( -1, 1 )

            avg = ValueError.average( data )
            if self.avg is None :
                self.avg = avg
        else :
            avg = self.avg

        return avg

    def punch( self, utm, radius ) :
        mask = self.circle_mask( utm, radius )

        if self.usenan :
            self.data[mask] = float( 'NaN' )
        else :
            self.data[mask] = self.nodata

        self.countgood -= mask[True].size
        self.ndmask[mask] = True

        used = self.data[~self.ndmask].reshape( -1, 1 )
        self.minimum = used.min()
        self.maximum = used.max()

        return self

    def circle_mask( self, utm, radius ) :
        mask = np.array( [ False ] * self.data.size ).reshape( self.data.shape )
        px = self.utm2pixel( utm )

        pr = radius / self.cellsize
        ir = int(pr) + 1

        lx = int(px.pixel[0]) - ir - 1
        tx = int(px.pixel[0]) + ir + 2
        ly = int(px.pixel[1]) - ir - 1
        ty = int(px.pixel[1]) + ir + 2

        if lx < 0 :
            lx = 0
        if tx > self.columns :
            tx = self.columns

        if ly < 0 :
            ly = 0
        if ty > self.rows :
            ty = self.rows

        xc = px.pixel[0]
        yc = self.rows - px.pixel[1]
        for x in range(lx, tx) :
            for y in range(ly, ty) :
                y = self.rows - y

                if (self.usenan and np.isnan(self.data[y, x])) or (self.data[y, x] < self.minimum) :
                    continue

                dx = x - xc
                dy = y - yc
                d = math.sqrt( (dx * dx) + (dy * dy) )

                mask[y, x] = d <= pr

        return mask

    def size( self ) :
        return self.rows * self.columns

    def cellarea( self ) :
        return self.cellsize * self.cellsize

    def surface( self ) :
        return self.cellsize * self.cellsize * self.rows * self.columns

    def used_surface( self ) :
        return self.cellsize * self.cellsize * self.countgood

    def height( self ) :
        return self.cellsize * self.rows

    def width( self ) :
        return self.cellsize * self.columns

    def pixel2meters( self, px ) :
        return self.cellsize * px

    def pixel2utm( self, px, revy=False ) :
        if (self.xllcorner is not None) and (self.yllcorner is not None) and (self.cellsize is not None) and (self.rows is not None) :
            x = px.pixel[0]
            if revy :
                y = self.rows - px.pixel[1]
            else :
                y = px.pixel[1]
                
            utm_x = self.xllcorner + (self.cellsize * x)
            utm_y = self.yllcorner + (self.cellsize * y)

            res = Pixel( utm_x, utm_y, px.value )
        else :
            res = px

        return res

    def utm2pixel( self, utm ) :
        if (self.xllcorner is not None) and (self.yllcorner is not None) and (self.cellsize is not None) and (self.rows is not None) :
            px = (utm[0] - self.xllcorner) / self.cellsize
            py = (utm[1] - self.yllcorner) / self.cellsize

            res = Pixel( px, py )
        else :
            res = Pixel( utm[0], utm[1] )

        return res

    def plot( self, **kwargs ) :
        figure = self.draw( **kwargs )
        plt.show()

        return figure

    def draw( self, filename=None, draw=True, title=None, hmask=None ) :
        if not draw and filename is None :
            return self

        formatter = tckr.FuncFormatter( lambda v, pos: "{:.0f}".format(v) )
        figure, axis = plt.subplots( 1, 1, figsize=self.figsize )

        if self.shadow is not None :
            plt.imshow( self.shadow, cmap=plt.cm.gist_gray, vmin=self.minimum, vmax=self.maximum, extent=self.extent )

        main = plt.imshow( self.data, cmap=self.colormap, vmin=self.minimum, vmax=self.maximum, extent=self.extent )
        if hmask is not None :
            core = self.data[hmask]

            if core.size > 0 :
                masked = np.array( self.data )
                masked[~hmask] = float('NaN')
            
                plt.imshow( masked, cmap=plt.cm.rainbow, vmin=core.min(), vmax=core.max(), extent=self.extent )

        axis.xaxis.set_major_formatter( formatter )
        axis.yaxis.set_major_formatter( formatter )

        if AsciiGrid.large :
            plt.xlabel( "Easting (m)", size=20 )
            plt.ylabel( "Northing (m)", size=20 )
            axis.tick_params( axis='both', labelsize=18 )
            axis.xaxis.set_major_locator( plt.MaxNLocator(5) )
        else :
            plt.xlabel( "Easting (m)" )
            plt.ylabel( "Northing (m)" )

        if self.usebar :
            cbar = plt.colorbar( main )

            if AsciiGrid.large :
                cbar.set_label( "T (°C)", rotation=270, labelpad=15, size=18 )
                cbar.ax.tick_params( labelsize=18 )
            else :
                cbar.set_label( "T (°C)", rotation=270, labelpad=15 )

        if not AsciiGrid.large :
            if title is None :
                figure.suptitle( self.filename )
            elif title != "" :
                figure.suptitle( title )

        if draw :
            plt.draw()

        if filename is not None :
            if AsciiGrid.large :
                figure.savefig( filename, dpi=300, bbox_inches="tight" )
            else :
                figure.savefig( filename )
            plt.close( figure )

            if AsciiGrid.verbose :
                logger.log( "Written image file <{}>".format(filename) )

        self.axis = axis

        return figure # self

    def silhouette( self ) :
        mask = np.array( [ 1 ] * self.data.size ).reshape( self.data.shape )
        mask[self.ndmask] = 0

        return mask

    def draw_both( self, other, alpha ) :
        if self.minimum < other.minimum :
            minimum = self.minimum
        else :
            minimum = other.minimum

        if self.maximum > other.maximum :
            maximum = self.maximum
        else :
            maximum = other.maximum

        formatter = tckr.FuncFormatter( lambda v, pos: "{:.0f}".format(v) )
        figure, axis = plt.subplots( 1, 1, figsize=self.figsize )

        plt.imshow( self.data, cmap=self.colormap, vmin=minimum, vmax=maximum, extent=self.extent )
        plt.imshow( other.data, cmap=self.colormap, vmin=minimum, vmax=maximum, extent=other.extent, alpha=alpha )
        axis.xaxis.set_major_formatter( formatter )
        axis.yaxis.set_major_formatter( formatter )

        if AsciiGrid.large :
            plt.xlabel( "Easting (m)", size=20 )
            plt.ylabel( "Northing (m)", size=20 )
            axis.tick_params( axis='both', labelsize=18 )
            axis.xaxis.set_major_locator( plt.MaxNLocator(5) )
        else :
            plt.xlabel( "Easting (m)" )
            plt.ylabel( "Northing (m)" )

        cbar = plt.colorbar()
        if AsciiGrid.large :
            cbar.set_label( "T (°C)", rotation=270, labelpad=15, size=18 )
        else :
            cbar.set_label( "T (°C)", rotation=270, labelpad=15 )

        plt.show()

        return self

class GridWriter :
    good = False
    columns = rows = None
    xllcorner = yllcorner = cellsize = None
    nodata = None
    
    def __init__( self, grid ) :
        if isinstance(grid, AsciiGrid) :
            self.good = grid.good

            if self.good :
                self.columns = grid.columns
                self.rows    = grid.rows

                self.xllcorner = grid.xllcorner
                self.yllcorner = grid.yllcorner
                self.cellsize  = grid.cellsize

                self.nodata = grid.nodata

        return

    def _write( self, handle, name, value ) :
        line = "{}\t{}".format( name, value )
        print( line.replace('.', ',').replace(",0 ", " "), file=handle )

        return

    def write_image( self, image, filename ) :
        with open( filename, "w" ) as handle :
            self._write( handle, "ncols", self.columns )
            self._write( handle, "nrows", self.rows )
            self._write( handle, "xllcorner", self.xllcorner )
            self._write( handle, "yllcorner", self.yllcorner )
            self._write( handle, "cellsize", self.cellsize )

            if self.nodata == int(self.nodata) :
                self._write( handle, "NODATA_value", int(self.nodata) )
            else :
                self._write( handle, "NODATA_value", self.nodata )

            for l in image :
                if self.nodata == int(self.nodata) :
                    line = " ".join( [ str(x) if x != self.nodata else str(int(x)) for x in l ] )
                else :
                    line = " ".join( [ str(x) for x in l ] )

                line = line.replace( '.', ',' ).replace( ",0 ", " " )
                print( line, file=handle )

        return self
