#!/usr/bin/python
#  -*- coding: utf-8 -*-

import sys
import pprint
import math
import copy

import statistics
import sklearn.cluster
import skimage.filters
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as tckr
import matplotlib.colors as colors

from collections import Counter

import logger
from utils import ValueError, Pixel

this = sys.modules[__name__]
this.verbose = False
this.draw = True
this.title = None
this.fluxmap = None
this.large = False
    
def set_title( t=None ) :
    this.title = t

    return

def set_large( l=True ) :
    this.large = l

    return

def set_verbose( v=True ) :
    this.verbose = v

    return

def enable_draw( d=True ) :
    this.draw = d

    return

def utm_axis( axis ) :
    formatter = tckr.FuncFormatter( lambda v, pos: "{:.0f}".format(v) )

    axis.xaxis.set_major_formatter( formatter )
    axis.yaxis.set_major_formatter( formatter )

    if this.large :
        axis.set_xlabel( "Easting (m)", size=20 )
        axis.set_ylabel( "Northing (m)", size=20 )
        axis.tick_params( axis='both', labelsize=18 )
        axis.xaxis.set_major_locator( plt.MaxNLocator(5) )
    else :
        axis.set_xlabel( "Easting (m)" )
        axis.set_ylabel( "Northing (m)" )

    return

def draw_figure( figure, filename ) :
    if this.draw :
        plt.draw()

    if filename is not None :
        if this.large :
            figure.savefig( filename, dpi=300, bbox_inches="tight"  )
        else :
            figure.savefig( filename )
        plt.close( figure )

        if this.verbose :
            logger.log( "Written image file <{}>".format(filename) )

    return

def create_fluxmap() :
    # g -> y -> r
    colarr = [ (0, 1, 0), (1, 1, 0), (1, 0, 0) ]

    cmap = colors.LinearSegmentedColormap.from_list( 'flux', colarr, N=100 )

    return cmap

def best_geometry( number ) :
    shape = None

    if number == 1 :
        shape = (1, 1)
    elif number == 2 :
        shape = (1, 2)
    elif (number == 3) or (number == 4) :
        shape = (2, 2)
    elif (number == 5) or (number == 6) :
        shape = (2, 3)
    elif (number >= 7) and (number <= 9) :
        shape = (3, 3)

    return shape

def identify_segments( segm ) :
    segments = segm.segments

    minval = minidx = None
    maxval = maxidx = None
    nodidx = None

    if segm.nans :
        noidx = segm.sortlabs[0][0]

        separated = {
            Clusterer.CLUSTERIZE_ALL:  [ lc[0] for lc in segm.sortlabs[1:] ],
            Clusterer.CLUSTERIZE_LOW:  [ segm.sortlabs[1][0] ],
            Clusterer.CLUSTERIZE_HIGH: [ segm.sortlabs[-1][0] ]
        }

        if segments <= 4 :
            separated[Clusterer.CLUSTERIZE_NOTLOW] = [ segm.sortlabs[1][0] ]
            separated[Clusterer.CLUSTERIZE_MID]    = [ lc[0] for lc in segm.sortlabs[1:-1] ]
        else :
            separated[Clusterer.CLUSTERIZE_NOTLOW] = [ segm.sortlabs[2][0] ]
            separated[Clusterer.CLUSTERIZE_MID]    = [ lc[0] for lc in segm.sortlabs[2:-1] ]
    else :
        separated = {
            Clusterer.CLUSTERIZE_ALL:   [ lc[0] for lc in segm.sortlabs ],
            Clusterer.CLUSTERIZE_LOW:   [ segm.sortlabs[0][0] ],
            Clusterer.CLUSTERIZE_HIGH:  [ segm.sortlabs[-1][0] ],
        }

        if segments <= 3 :
            separated[Clusterer.CLUSTERIZE_NOTLOW] = [ segm.sortlabs[0][0] ]
            separated[Clusterer.CLUSTERIZE_MID]    = [ lc[0] for lc in segm.sortlabs[0:-1] ]
        else :
            separated[Clusterer.CLUSTERIZE_NOTLOW] = [ segm.sortlabs[1][0] ]
            separated[Clusterer.CLUSTERIZE_MID]    = [ lc[0] for lc in segm.sortlabs[1:-1] ]

    minidx = separated[Clusterer.CLUSTERIZE_LOW][0]
    minval = segm.centers[minidx]

    maxidx = separated[Clusterer.CLUSTERIZE_HIGH][0]
    maxval = segm.centers[maxidx]
        
    return separated, minidx, minval, maxidx, maxval, nodidx

class Cluster :
    def __init__( self, seg_label, clus_label ) :
        self.segment = seg_label
        self.cluster = clus_label

        self.bary = ( None, None )
        self.meant = self.radius = None
        self.tmean = self.rmean = self.trms = self.drms = 0
        self.clusnum = 0

        self.points = self.clusters = self.expand = None

        return

    def __getitem__( self, pt ) :
        value = None
        pt = ( int(pt[0]), int(pt[1]) )
        if self.points and (pt in self.points) :
            value = self.points[pt]

        return value

    def recalculate( self ) :
        if self.points is None :
            return self
        
        length = len( self.points )

        if length == 0 :
            self.bary = ( None, None )
            self.meant = self.radius = None
            self.tmean = self.rmean = self.trms = self.drms = 0
            self.clusnum = 0
        elif length == 1 :
            p = next( iter(self.points.keys()) )
            self.bary = ( p[1], p[0] )
            self.tmean = self.points[p]

            self.rmean = 1
            self.trms = 0.1
            self.meant = ValueError( self.tmean, self.trms )

            self.drms = 0.1
            self.radius = ValueError( self.rmean, self.drms )
        else :            
            tsum = xsum = ysum = 0

            for p, t in self.points.items() :
                tsum += t
                xsum += t * p[1]
                ysum += t * p[0]

            self.tmean = tsum / length
            self.bary = ( xsum / tsum, ysum / tsum )

            dsum2 = tdist = 0
            dists = []
            for p, t in self.points.items() :
                dt = t - self.tmean
                dx = p[1] - self.bary[0]
                dy = p[0] - self.bary[1]
            
                dsum2 += dt * dt
                dd = math.sqrt( (dx * dx) + (dy * dy) + (dt * dt) ) # 3D distance with temperature

                tdist += dd
                dists.append( dd )

            self.rmean = tdist / length

            self.trms = math.sqrt( dsum2 / (length - 1) )
            self.meant = ValueError( self.tmean, self.trms )

            dsum2 = 0
            for d in dists :
                dd = d - self.rmean
                dsum2 += dd * dd

            self.drms = math.sqrt( dsum2 / (length - 1) )
            self.radius = ValueError( self.rmean, self.drms )

        return self

    def set_data( self, points, recalc=True, image=None ) :
        if image is None :
            self.points = { (int(p[0]), int(p[1])): p[2] for p in points }
        else :
            self.points = { (int(p[0]), int(p[1])): image[p[0], p[1]] for p in points }
        if recalc :
            self.recalculate()

        return self

    def append_data( self, points, recalc=True, image=None ) :
        if self.points is None :
            if image is None :
                self.points = { (int(p[0]), int(p[1])): p[2] for p in points }
            else :
                self.points = { (int(p[0]), int(p[1])): image[p[0], p[1]] for p in points }
        else :
            for p in points :
                pt = ( int(p[0]), int(p[1]) )
                self.points[pt] = p[2] if image is None else image[p[0], p[1]]

        if recalc :
            self.recalculate()

        return self

    def append_cluster_data( self, that, recalc=True, image=None ) :
        if self.points is None :
            self.points = copy.deepcopy( that.points )
        elif that.points is not None :
            for p, t in that.points.items() :
                self.points[p] = t if image is None else image[p[0], p[1]]

        if recalc :
            self.recalculate()

        return self

    def remove_border( self, points, dest=None, recalc=False, image=None ) :
        if self.points is None :
            return None

        removed = list()
        for p in points :
            pt = ( int(p[0]), int(p[1]) )
            t = self.points.pop( pt, None )

            if t is not None :
                removed.append( [ *pt, t ] )

        if removed :
            if recalc :
                self.recalculate()
            
            if dest is None :
                dest = Cluster( self.segment, self.cluster )
                dest.set_data( removed, recalc=recalc, image=image )
            else :
                dest.append_data( removed, recalc=recalc, image=image )

        return dest

    def set_clusters( self, clusters ) :
        points = None

        for c in clusters :
            if points is None :
                points = { p: t for p, t in c.points.items() }
            else :
                for p, t in c.points.items() :
                    points[p] = t

        if points :
            self.points = points
            self.clusters = copy.deepcopy( clusters )
            self.clusnum = len( clusters )

            self.recalculate()

        return self

    def to_pixel( self ) :
        return Pixel( *self.bary, self.tmean )

    def grow( self, pt=None ) :
        if pt is None :
            if self.expand is not None :
                if self.points is None :
                    self.points = { (int(p[0]), int(p[1])): p[2] for p in self.expand }
                else :
                    for p in self.expand :
                        pt = ( int(p[0]), int(p[1]) )
                        self.points[pt] = p[2]

                self.recalculate()
                self.expand = None
        elif self.expand is None :
            self.expand = [ pt ]
        else :
            self.expand.append( pt )

        return self

    def size( self ) :
        return 0 if self.points is None else len(self.points)

    def distance( self, x, y, t=None ) :
        if self.points is None :
            return None

        dist = None
        for p, tp in self.points.items() :
            dx = p[1] - x
            dy = p[0] - y

            if t is None :
                d = math.sqrt( (dx * dx) + (dy * dy) )
            else :
                dt = tp - t
                d = math.sqrt( (dx * dx) + (dy * dy) + (dt * dt) )

            if (dist is None) or (dist < d) :
                dist = d

        return dist

    def fill_image( self, image, mean=False, value=None ) :
        if self.points is None :
            return self
        
        for p, t in self.points.items() :
            if mean :
                image[p[0], p[1]] = self.tmean
            elif value is None :
                image[p[0], p[1]] = t
            else :
                image[p[0], p[1]] = value

        return self

    def draw_cluster( self, color ) :
        if self.points is None :
            return self
        
        pts = np.array( [ p for p in self.points.keys() ] )
        c = [ color ] * len(pts)

        plt.scatter( pts[:, 1], pts[:, 0], s=1, c=c, marker=',' )
        plt.scatter( *self.bary, s=20, c='black', marker='+' )

        return self

class ClusterDrawer :
    def __init__( self, segment, clusters, bg=None ) :
        self.segm = segment
        self.clusters = clusters
        self.background = bg
        
        self.title = "Cluster shapes"
        self.unique_labels = self.number = self.figshape = self.figsize = None

        return

    def set_background( self, image ) :
        self.background = image

        return self

    def set_title( self, title ) :
        self.title = title

        return self

    def subfigure_number( self ) :
        if self.unique_labels is None :
            self.unique_labels = [ lbl for lbl in self.segm.unique if (lbl in self.clusters) and (len(self.clusters[lbl]) != 0) ]
            number = len( self.unique_labels )

            if number > 1 :
                self.figshape = best_geometry( number )
                self.figsize = tuple( e1 * e2 for e1, e2 in zip(self.figshape, self.segm.figsize) )
            else :
                self.figshape = None
                self.figsize = self.segm.figsize

        return (self.unique_labels, self.figsize, self.figshape)

    def suptitle( self, figure, number ) :
        if this.title is None :
            if number > 1 :
                figure.suptitle( "{} for each segment".format(self.title) )
            else :
                figure.suptitle( self.title )
        elif number > 1 :
            figure.suptitle( "{}\n{} for each segment".format(this.title, self.title) )
        else :
            figure.suptitle( "{}\n{}".format(this.title, self.title) )

        return self
    
    def draw_clusters( self, filename, mean=True, value=None, borders=None ) :
        (unique_labels, figsize, figshape) = self.subfigure_number()
        number = len( unique_labels )

        figure = plt.figure( figsize=figsize )

        for segi in range(number) :
            lbl = unique_labels[segi]
            if self.segm.centers[lbl][0] < self.segm.minval :
                continue

            if number > 1 :
                ax = figure.add_subplot( figshape[0], figshape[1], segi + 1 ) # 1 to n not 0 to n-1
                ax.set_title( "Image segment {}: {:.1f}°C".format(lbl, self.segm.centers[lbl][0]) )

                if self.background is not None :
                    plt.imshow( self.background, vmin=self.segm.minval, vmax=self.segm.maxval, cmap=plt.cm.gist_gray, extent=self.segm.extent )

                    if self.segm.extent is not None :
                        utm_axis( ax )
            else :
                if not this.large :
                    plt.title( "Image segment {}: {:.1f}°C".format(lbl, self.segm.centers[lbl][0]) )

                if self.background is not None :
                    plt.imshow( self.background, vmin=self.segm.minval, vmax=self.segm.maxval, cmap=plt.cm.gist_gray, extent=self.segm.extent )

                    if self.segm.extent is not None :
                        utm_axis( plt.gca() )

            clsimage = np.array( [ float('NaN') ] * self.segm.size ).reshape( self.segm.shape )
            for c in self.clusters[lbl] :
                c.fill_image( clsimage, mean=mean, value=value )

            if borders is not None :
                for c in borders[lbl] :
                    c.fill_image( clsimage, value=self.segm.maxval )

            plt.imshow( clsimage, cmap=plt.cm.inferno, vmin=self.segm.minval, vmax=self.segm.maxval, extent=self.segm.extent )

        if not this.large :
            self.suptitle( figure, number )
        draw_figure( figure, filename )

        return self

    def draw_centroids( self, grid, filename ) :
        (unique_labels, figsize, figshape) = self.subfigure_number()
        number = len( unique_labels )
  
        figure = plt.figure( figsize=figsize )
        norm = colors.Normalize( vmin=self.segm.minval, vmax=self.segm.maxval )
        mapper = plt.cm.ScalarMappable( norm=norm, cmap=plt.cm.inferno ) # cmap=self.segm.colormap )

        for segi in range(number) :
            lbl = unique_labels[segi]
            if self.segm.centers[lbl][0] < self.segm.minval :
                continue
            
            if number > 1 :
                ax = figure.add_subplot( figshape[0], figshape[1], segi + 1 )
                ax.set_title( "Image segment {}: {:.1f}°C".format(lbl, self.segm.centers[lbl][0]) )

                if self.background is not None :
                    plt.imshow( self.background, vmin=self.segm.minval, vmax=self.segm.maxval, cmap=plt.cm.gist_gray, extent=self.segm.extent )

                    if self.segm.extent is not None :
                        utm_axis( ax )
            else :
                if not this.large :
                    plt.title( "Image segment {}: {:.1f}°C".format(lbl, self.segm.centers[lbl][0]) )

                if self.background is not None :
                    plt.imshow( self.background, vmin=self.segm.minval, vmax=self.segm.maxval, cmap=plt.cm.gist_gray, extent=self.segm.extent )

                    if self.segm.extent is not None :
                        utm_axis( plt.gca() )

            x = []
            y = []
            c = []
            s = []

            for cl in self.clusters[lbl] :
                utm = grid.pixel2utm( cl.to_pixel(), True )
                radius = grid.pixel2meters( cl.rmean )

                x.append( utm.pixel[0] )
                y.append( utm.pixel[1] )
                s.append( math.pi * radius * radius )
                c.append( mapper.to_rgba(cl.tmean) )

            plt.scatter( x, y, s=s, c=c, lw=1, edgecolors='black' )

        if not this.large :
            self.suptitle( figure, number )
        draw_figure( figure, filename )
        
        return self

#    def draw_centroid( self, color ) :
#        radius = 20 * self.rmean
#        plt.scatter( self.bary[0], self.bary[1], s=radius, c=[color], marker='o', linewidths=1, edgecolors='black' )
#
#        return self
        
class Segmenter :
    figsize = (9, 5)
    colormap = plt.cm.gnuplot2
    nans = False
    nanval = None

    labels = image = segments = None
    sortlabs = None

    centers = None
    
    def __init__( self, data, minval, maxval, extent=None ) :
        self.shape = data.shape
        self.size = data.size
        self.extent = extent

        tmp = np.array( data )
        nanp = np.isnan( tmp )
        self.nans = nanp.any()
        self.nanval = minval - 100

        if self.nans :
            tmp[nanp] = self.nanval

        self.data = np.float32( tmp ).reshape( (-1, 1) )

        self.minval = minval
        self.maxval = maxval

        self.shadow = None

        return

    def use_shadow( self, shadow ) :
        self.shadow = shadow

        return self

    def use_size( self, size ) :
        self.figsize = size

        return self

    def use_colormap( self, cmap ) :
        self.colormap = cmap

        return self

    def _show_segments( self, loop=None ) :
        #if loop is not None :
        #    logger.log( "Pass {:02d} ->".format(loop), end=' ' )
        logger.log( "Image quantized in {} segments:".format(len(self.centers)) )

        for i in range(len(self.sortlabs)) :
            lc = self.sortlabs[i]
            j = lc[0]

            if self.nans and (i == 0) :
                logger.log( "|-> Segment {}: NODATA, {} pixels".format(lc[0], self.sizes[j]) )
            else :
                logger.log( "|-> Segment {}: {:.1f} °C, {} pixels".format(lc[0], lc[1], self.sizes[j]) )

        return

    def _sorted_labels( self ) :
        sortlabs = [ ( i, self.centers[i][0] ) for i in range(self.segments) ]
        sortlabs.sort( key = lambda pt: pt[1] )

        return sortlabs

    def minimum_segment( self ) :
        if self.nans :
            idx = 1
        else :
            idx = 0

        answer = ( idx, *self.sortlabs[idx] )

        return answer

    def segment_by_index( self, idx ) :
        if idx <= len(self.sortlabs) :
            answer = self.sortlabs[idx]
        else :
            answer = None

        return answer

    def quite_minimum_segment( self ) :
        if self.nans :
            if self.segments <= 4 :
                minimum = self.minimum_segment()
            else :
                minimum = self.sortlabs[2]
        elif self.segments <= 3 :
            minimum = self.minimum_segment()
        else :
            minimum = self.sortlabs[1]

        return minimum

    #
    # Segment data into 'number' segments
    #
    def segment( self, number, preseed=False, loop=None ) :
        if this.verbose :
            logger.log( "Segmenting start", end='', flush=True )

        if self.nans :
            # Add a segment for NODATA
            number += 1

            if this.verbose :
                logger.log( ", added NODATA segment", end='', flush=True )

        if preseed :
            if self.nans :
                nints = number - 1
            else :
                nints = number
                
            width = (self.maxval - self.minval) / nints

            start = self.minval + (width / 2)
            end   = self.maxval - (width / 2)
            centers = np.linspace( start, end, nints )

            if loop is None :
                ctr = 1
            else :
                ctr = loop - 1

            samples = self.data[self.data >= self.minval].flatten()
            try :
                centers[ctr] = statistics.mode( samples ) # Try to segment around the mode
            except statistics.StatisticsError as err :
                if this.verbose :
                    logger.log( "... Failed!" )
                    logger.log( "Cannot extract data mode (multimodal distribution?): \"{}\"".format(err) )
                    logger.log( "Using mean value, instead; segmenting restarted", end='', flush=True )

                centers[ctr] = samples.sum() / samples.size

            if self.nans :
                centers = np.array( [ self.nanval ] + [ x for x in centers] ).reshape( -1, 1 )
            else :
                centers = centers.reshape( -1, 1 )

            if this.verbose :
                logger.log( ", initialized with preseeded centers: [ {} ]...".format(", ".join(["{:.1f}".format(c[0]) for c in centers])), end=' ', flush=True )

            kmeans = sklearn.cluster.KMeans( n_clusters=number, init=centers, n_init=1, random_state=0 ).fit( self.data )
        else :
            if this.verbose :
                logger.log( "...", end=' ', flush=True )

            kmeans = sklearn.cluster.KMeans( n_clusters=number, n_init=10, random_state=0 ).fit( self.data )

        labels = kmeans.predict( self.data )

        self.centers = kmeans.cluster_centers_
        self.labels = labels.flatten()
        self.unique = set( self.labels )
        self.image = self.centers[self.labels].reshape( self.shape )
        self.sizes = Counter( self.labels )

        self.segments = number
        self.sortlabs = self._sorted_labels()

        if this.verbose :
            logger.log( "Done!" )
            self._show_segments( loop )

        return self

    #
    # Segment over threshold
    #
    def threshold( self, value ) :
        number = 2 # Index of second label 0, 1
        
        if self.nans :
            # Add a segment for NODATA
            number += 1

            if this.verbose :
                logger.log( "Threshold segmenting start ({:.1f} °C), added NODATA segment...".format(value), end='', flush=True )
        elif this.verbose :
            logger.log( "Threshold segmenting start ({:.1f} °C)...".format(value), end='', flush=True )

        labels = np.array( [ 1 ] * len(self.data) ).reshape( (-1, 1) )
        labels[self.data < value] = 0
        if self.nans :
            labels[self.data < self.minval] = 2

        self.labels = labels.flatten()

        self.sizes = Counter( self.labels )
        self.unique = set( self.labels )
        self.centers = np.array( [ self.data[self.labels == i].sum() / self.sizes[i] for i in range(number) ] ).reshape( (-1, 1) )
        self.image = self.centers[self.labels].reshape( self.shape )

        self.segments = number
        self.sortlabs = self._sorted_labels()

        if this.verbose :
            self._show_segments()

        return self

    #
    # Segment on mask
    #
    def apply_mask( self, mask ) :
        number = 2 # Index of second label 0, 1

        if self.nans :
            # Add a NODATA segment
            number += 1

            if this.verbose :
                logger.log( "Mask segmenting start, added NODATA segment", end=' ', flush=True )
        elif this.verbose :
            logger.log( "Mask segmenting start...", end=' ', flush=True )

        flatmask = np.array( mask ).reshape( (-1, 1) )
        labels = np.array( [ number - 2 ] * len(self.data) ).reshape( (-1, 1) )
        labels[flatmask] = number - 1
        if self.nans :
            labels[self.data < self.minval] = number - 3

        self.labels = labels.flatten()

        self.sizes = Counter( self.labels )
        self.unique = set( self.labels )
        self.centers = np.array( [ self.data[self.labels == i].sum() / self.sizes[i] for i in range(number) ] ).reshape( (-1, 1) )
        self.image = self.centers[self.labels].reshape( self.shape )

        self.segments = number
        self.sortlabs = self._sorted_labels()

        if this.verbose :
            self._show_segments()

        return self

    def draw_segments( self, filename=None, title=None ) :
        if not this.draw and filename is None :
            return self
            
        figure, axis = plt.subplots( 1, 1, figsize=self.figsize )

        if self.shadow is not None :
            shadow = np.array( self.shadow )
            shadow[shadow < self.minval] = float( 'NaN' )

            plt.imshow( shadow, cmap=plt.cm.gist_gray, vmin=self.minval, vmax=self.maxval, extent=self.extent )

        data = self.image
        data[data < self.minval] = float( 'NaN' )

        plt.imshow( data, interpolation='nearest', cmap=self.colormap, vmin=self.minval, vmax=self.maxval, extent=self.extent )
        if self.extent is not None :
            utm_axis( axis )

        cbar = plt.colorbar()
        if this.large :
            cbar.set_label( "T (°C)", rotation=270, labelpad=15, size=18 )
            cbar.ax.tick_params( labelsize=18 )
        else :
            cbar.set_label( "T (°C)", rotation=270, labelpad=15 )

        if not this.large :
            if title is None :
                title = "K-MEANS segmentation ({})".format(self.segments)
            
            if this.title is None :
                figure.suptitle( title )
            else :
                figure.suptitle( "{}\n{}".format(this.title, title) )

        draw_figure( figure, filename )

        return self

    def draw_histogram( self, filename=None, title=None ) :
        if not this.draw and filename is None :
            return self
        
        datasets = dict()
        for i in range(len(self.labels)) :
            lbl = self.labels[i]

            if self.centers[lbl][0] >= self.minval :
                if lbl in datasets :
                    datasets[lbl].append( self.data[i] )
                else :
                    datasets[lbl] = [ self.data[i] ]

        norm = colors.Normalize( vmin=self.minval, vmax=self.maxval )
        mapper = plt.cm.ScalarMappable( norm=norm, cmap=self.colormap )

        histos = dict()
        for lbl in datasets :
            minval = min( datasets[lbl] )
            maxval = max( datasets[lbl] )
            n = int( maxval - minval ) + 1

            histo, edges = np.histogram( datasets[lbl], n )

            histos[lbl] = { 'bins': n,
                            'data': histo,
                            'edges': edges,
                            'color': mapper.to_rgba(self.centers[lbl]) }

        figure = plt.figure( figsize=self.figsize )

        for lbl in histos :
            hc = (histos[lbl]['edges'][:-1] + histos[lbl]['edges'][1:]) / 2
            width = np.diff( histos[lbl]['edges'] )

            plt.bar( hc, histos[lbl]['data'],
                     align="center", width=width, color=histos[lbl]['color'],
                     edgecolor="black", label="{:.1f}°C".format(self.centers[lbl][0]) )

        plt.legend()

        if title is None :
            title = "K-MEANS segmented temperature distribution ({})".format(self.segments)
            
        if this.title is None :
            plt.title( title )
        else :
            plt.title( "{}\n{}".format(this.title, title) )

        plt.xlabel( "Temperatures (°C)" )
        plt.ylabel( "Pixel number" )
        plt.grid()

        draw_figure( figure, filename )

        return self

    def draw_fluximage( self, image, filename=None, fluxmap=None, moretitle=None ) :
        if not this.draw and filename is None :
            return self

        if fluxmap is None :
            if this.fluxmap is None :
                this.fluxmap = create_fluxmap()

            fluxmap = this.fluxmap
            
        figure, axis = plt.subplots( 1, 1, figsize=self.figsize )        

        plt.imshow( image.flux, cmap=fluxmap, vmin=image.minimum, vmax=image.maximum, extent=self.extent )
        if self.extent is not None :
            utm_axis( axis )

        cbar = plt.colorbar()
        if this.large :
            cbar.set_label( "Heat flux (W/m$^2$)", rotation=270, labelpad=15, size=18 )
            cbar.ax.tick_params( labelsize=18 )
        else :
            cbar.set_label( "Heat flux (W/m$^2$)", rotation=270, labelpad=15 )

        if not this.large :
            if this.title is None :
                if moretitle is None :
                    suptitle = "Heat flux"
                else :
                    suptitle = "Heat flux ({})".format(moretitle)
            else :
                if moretitle is None :
                    suptitle = "{}\nHeat flux".format(this.title)
                else :
                    suptitle = "{}\nHeat flux ({})".format(this.title, moretitle)
            figure.suptitle( suptitle )

        draw_figure( figure, filename )

        return

class Enhancer :
    ENHANCE_ALL    = -1
    ENHANCE_LOW    = -2
    ENHANCE_NOTLOW = -3
    ENHANCE_MID    = -4
    ENHANCE_HIGH   = -5
    ENHANCE_BOTH   = -6

    def __init__( self, segm ) :
        self.good = isinstance( segm, Segmenter )

        if not self.good :
            return

        self.segm = segm
        self.data = None

    @classmethod
    def string_mode( cls, mode ) :
        if mode == Enhancer.ENHANCE_ALL :
            mode = "all"
        elif mode == Enhancer.ENHANCE_LOW :
            mode = "low"
        elif mode == Enhancer.ENHANCE_NOTLOW :
            mode = "quite low"
        elif mode == Enhancer.ENHANCE_MID :
            mode = "mid"
        elif mode == Enhancer.ENHANCE_HIGH :
            mode = "high"
        elif mode == Enhancer.ENHANCE_BOTH :
            mode = "highest and brother"
        else :
            mode = "unkn"

        return mode

    def enhance( self, height, loss, mode=None ) :
        if not self.good :
            return self

        total = height * loss
        if this.verbose :
            logger.log( "Enhancing image temperature. Flight height is {}m, temperature loss is {:.2f}%/m, total delta is: {:.2f}%".format(height, loss, total), end='' )

        if mode is None :
            mode = Enhancer.ENHANCE_HIGH

        if mode == Enhancer.ENHANCE_BOTH :
            self.separated = { mode: [ self.segm.sortlabs[-2][0], self.segm.sortlabs[-1][0] ] }

            if this.verbose :
                logger.log( ", mode: \"{}\" ({})".format(Enhancer.string_mode(mode), self.separated[mode]) )

        elif mode >= 0 :
            indexes = [ index for index, value in enumerate(self.segm.sortlabs) if value[0] == mode ]
            self.separated = { indexes[0]: [ mode ] }

            if this.verbose :
                logger.log( ", requested only segment {} ({}).".format(mode, indexes[0]) )

        else :
            self.separated, self.minidx, self.minval, self.maxidx, self.maxval, self.noidx = identify_segments( self.segm )
            if mode == Clusterer.CLUSTERIZE_NOTLOW :
                self.bgidx = self.separated[Clusterer.CLUSTERIZE_NOTLOW][0]
            else :
                self.bgidx = self.minidx

            if this.verbose :
                logger.log( ", mode: \"{}\"".format(Enhancer.string_mode(mode)) )

        for lbl in self.separated[mode] :
            if this.verbose :
                logger.log( "Enhancing segment {}: {:.1f}°C".format(lbl, self.segm.centers[lbl][0]), end='' )

            data = np.array( self.segm.data ).reshape( (-1, 1) )
            data[self.segm.labels == lbl] *= (1 + total/100)

            self.minval = min( data[data > self.segm.nanval] )
            self.maxval = max( data[data > self.segm.nanval] )

            self.segm.data = data
            self.segm.centers[lbl][0] = float( data[self.segm.labels == lbl].sum() / self.segm.sizes[lbl] )

            for i in range(len(self.segm.sortlabs)) :
                lc = self.segm.sortlabs[i]
                j = lc[0]

                if j == lbl :
                    self.segm.sortlabs[i] = tuple( [j, self.segm.centers[lbl][0]] )

            self.data = np.array( data ).reshape( self.segm.shape )

            if this.verbose :
                logger.log( ", enhanced at {:.1f}°C".format(self.segm.centers[lbl][0]) );

        return self

    def draw_enhanced( self, filename=None, title=None ) :
        if not this.draw and filename is None :
            return self

        figure, axis = plt.subplots( 1, 1, figsize=self.segm.figsize )

        data = np.array( self.data ).reshape( self.segm.shape )
        data[data < self.minval] = float( 'NaN' )

        plt.imshow( data, interpolation="nearest", cmap=self.segm.colormap, vmin=self.minval, vmax=self.maxval, extent=self.segm.extent )
        if self.segm.extent is not None :
            utm_axis( axis )

        cbar = plt.colorbar()
        if this.large :
            cbar.set_label( "T (°C)", rotation=270, labelpad=15, size=18 )
            cbar.ax.tick_params( labelsize=18 )
        else :
            cbar.set_label( "T (°C)", rotation=270, labelpad=15 )

        if not this.large :
            if title is None :
                title = "Enhanced temperature map"

            if this.title is None :
                figure.suptitle( title )
            else :
                figure.suptitle( "{}\n{}".format(this.title, title) )

        draw_figure( figure, filename )

        return self

class Clusterer :
    CLUSTERIZE_ALL    = -1
    CLUSTERIZE_LOW    = -2
    CLUSTERIZE_NOTLOW = -3
    CLUSTERIZE_MID    = -4
    CLUSTERIZE_HIGH   = -5

    def __init__( self, segm ) :
        self.good = isinstance( segm, Segmenter )
        if not self.good :
            return
        
        self.segm = segm
        self.segcrd = self.segments = self.clusters = self.noises = None
        self.sprclus = self.sprnois = self.borders = None
        self.powerclus = None
        self.minidx = self.maxidx = self.minval = self.maxval = self.bgidx = self.nodidx = self.separated = None

        labels = segm.labels.reshape( segm.shape )
        self.segcrd = dict()
        for lbl in range(segm.segments) :
            self.segcrd[lbl] = np.transpose( np.where(labels==lbl) )

        data = segm.data.reshape( segm.shape )
        self.segments = { lbl: [] for lbl in range(segm.segments) }
        for lbl in range(segm.segments) :
            for pt in self.segcrd[lbl] :
                self.segments[lbl].append( [ pt[0], pt[1], data[pt[0]][pt[1]] ] )

        return

    @classmethod
    def string_mode( cls, mode ) :
        if mode == Clusterer.CLUSTERIZE_ALL :
            mode = "all"
        elif mode == Clusterer.CLUSTERIZE_LOW :
            mode = "low"
        elif mode == Clusterer.CLUSTERIZE_NOTLOW :
            mode = "quite low"
        elif mode == Clusterer.CLUSTERIZE_MID :
            mode = "mid"
        elif mode == Clusterer.CLUSTERIZE_HIGH :
            mode = "high"
        else :
            mode = "unkn"

        return mode

    def get_thermal_clusters( self, lbl=None ) :
        if (lbl is not None) and (lbl in self.clusters) :
            result = self.clusters[lbl]
        elif (self.maxidx is not None) and (self.maxidx in self.clusters) :
            result = self.clusters[self.maxidx]
        else :
            result = self.clusters

        return result

    def get_flux_clusters( self, lbl=None ) :
        if (lbl is not None) and (lbl in self.powerclus) :
            result = self.powerclus[lbl]
        elif (self.maxidx in self.powerclus) and (self.maxidx in self.powerclus) :
            result = self.powerclus[self.maxidx]
        else :
            result = self.powerclus

        return result


    def clusterize( self, distance, minpt, mode=None ) :
        if not self.good :
            return self

        if this.verbose :
            logger.log( "Clustering start", end='' )

        if mode is None :
            mode = Clusterer.CLUSTERIZE_ALL

        if mode >= 0 :
            indexes = [ index for index, value in enumerate(self.segm.sortlabs) if value[0] == mode ]
            self.separated = { indexes[0]: [ mode ] }

            if this.verbose :
                logger.log( ", requested only segment {} ({})".format(mode, indexes[0]), end='' )

            mode = indexes[0]
        else :
            self.separated, self.minidx, self.minval, self.maxidx, self.maxval, self.noidx = identify_segments( self.segm )
            if mode == Clusterer.CLUSTERIZE_NOTLOW :
                self.bgidx = self.separated[Clusterer.CLUSTERIZE_NOTLOW][0]
            else :
                self.bgidx = self.minidx

            if this.verbose :
                logger.log( ", mode: \"{}\"".format(Clusterer.string_mode(mode)), end='' )

        if this.verbose :
            logger.log( ", distance: {}, minimum points: {}...".format(distance, minpt) )

        self.clusters = dict()
        self.noises = dict()
        for lbl in self.separated[mode] :
            if this.verbose :
                logger.log( "Clustering segment {}: {:.1f}°C...".format(lbl, self.segm.centers[lbl][0]), end='', flush=True )

            self.clusters[lbl] = []
            self.noises[lbl] = []

            S = np.array( self.segments[lbl] )
            labels = sklearn.cluster.DBSCAN( eps=distance, min_samples=minpt ).fit_predict( S )

            unique_labels = set( labels )
            clusnum = len( unique_labels )
            noisenum = 0
            if -1 in labels :
                clusnum -= 1
                noisenum = list( labels ).count( -1 )

            if this.verbose :
                logger.log( " Done! Found {} clusters over {} pixels, {} noise...".format(clusnum, len(self.segments[lbl]), noisenum) )
                logger.log( "Gathering clusters..." )

            for cn in unique_labels :
                if cn == -1 :
                    if this.verbose :
                        logger.log( "Skipping noise cluster" )

                    indices = [ i[0] for i in enumerate(labels) if i[1] == -1 ]
                    self.noises[lbl] = [ [int(pt[0]), int(pt[1]), pt[2]] for pt in S[indices] ]
                else :
                    indices = [ i[0] for i in enumerate(labels) if i[1] == cn ]

                    c = Cluster( lbl, cn )
                    c.set_data( S[indices] )

                    self.clusters[lbl].append( c )

            self.clusters[lbl].sort( key=lambda c: c.size(), reverse=True )
            if this.verbose :
                for c in self.clusters[lbl] :
                    logger.log( "Grouped cluster {:02d}:{:03d} => T = {}, center = ({:.1f}, {:.1f}), radius = {}, size: {}".
                           format(lbl, c.cluster, c.meant, c.bary[0], c.bary[1], c.radius, c.size()) )

        return self

    def extract_background( self, frac=0.75, segment=None ) :
        if not self.good or ((self.bgidx is None) and (segment is None)) :
            return None

        if segment is None :
            minidx = self.bgidx
        else :
            minidx = segment

        if this.verbose :
            logger.log( "Minimum temperature segment is {} ({:.1f} °C)".format(minidx, self.segm.centers[minidx][0]) )

        segsize = len( self.segments[minidx] )
        minsize = int( frac * segsize )
        if this.verbose :
            logger.log( "Minimum number of background pixels: {} over {}".format(minsize, segsize) )

        totsize = 0;
        clustemp = []
        for c in self.clusters[minidx] :
            totsize += c.size()
            clustemp.append( ValueError(c.tmean, c.trms, c.size()) )

            if this.verbose :
                logger.log( "Cluster {:02d}:{:03d}, size {}, total: {}/{} ({:.1f} %)".format(c.segment, c.cluster, c.size(), totsize, minsize, 100. * totsize / minsize) )

            if totsize >= minsize :
                break

        if len(clustemp) == 1 :
            Tbg = clustemp[0]
        else :
            Tbg = ValueError.weight_mean( clustemp )

        return Tbg

    def extreme_warp( self, temp=None ) :
        other = Clusterer( self.segm )

        if this.verbose :
            logger.log( "Warping middle temperature clusters into extremes..." )

        if (self.separated is None) and (self.minidx is None) and (self.minval is None) and (self.maxidx is None) and (self.maxval is None) and (self.nodidx is None) :
            separated, minidx, minval, maxidx, maxval, nodidx = identify_segments( self.segm )

            other.separated = separated
            other.maxidx = maxidx
            other.minidx = minidx
            other.minval = minval
            other.maxval = maxval
            other.nodidx = noidx
        else :
            other.separated = self.separated.copy()
            other.maxidx = self.maxidx
            other.maxval = self.maxval
            other.minidx = self.minidx
            other.minval = self.minval
            other.nodidx = self.nodidx

        other.clusters = dict()
        other.noises = dict()

        for lbl in self.segm.unique :
            other.clusters[lbl] = []
            other.noises[lbl] = []
            
            if temp is None :
                threshold = self.segm.centers[lbl][0]
            else :
                threshold = temp

            if this.verbose :
                logger.log( "Warping segment {}, threshold = {:.1f}°C".format(lbl, threshold) )

            if lbl == other.nodidx :
                if this.verbose :
                    logger.log( "Skipping NODATA segmente {}".format(lbl) )
                    
                continue
            elif (lbl == other.minidx) or (lbl == other.maxidx) :
                other.clusters[lbl].extend( self.clusters[lbl] )
                other.noises[lbl].extend( self.noises[lbl] )

                if this.verbose :
                    logger.log( "Stored extremes clusters and noises for segment {}".format(lbl) )
            else :
                for c in self.clusters[lbl] :
                    if c.tmean > threshold :
                        other.clusters[other.maxidx].append( c )

                        if this.verbose :
                            logger.log( "Warped cluster {}:{} ({:.1f}°C) to hot segment {}".format(c.segment, c.cluster, c.tmean, maxidx) )
                    else :
                        other.clusters[other.minidx].append( c )

                        if this.verbose :
                            logger.log( "Warped cluster {}:{} ({:.1f}°C) to cold segment {}".format(c.segment, c.cluster, c.tmean, minidx) )

                for pt in self.noises[lbl] :
                    if pt[2] > threshold :
                        other.noises[other.maxidx].append( pt )
                    else :
                        other.noises[other.minidx].append( pt )

        return other

    def _retrieve_labels( self, onlyhot ) :
        if onlyhot :
            if self.maxidx is not None :
                seglabels = [ self.maxidx ]
            else :
                maxtemp = maxidx = None
                
                for lbl in self.segm.unique :
                    if self.segm.centers[lbl][0] < self.segm.minval :
                        continue
                    elif (maxtemp is None) or (maxidx is None) or (maxtemp < self.segm.centers[lbl][0]) :
                        maxidx = lbl
                        maxtemp = self.segm.centers[lbl][0]

                self.maxidx = maxidx
                seglabels = [ maxidx ]
        else :
            seglabels = self.segm.unique

        return seglabels

    def reclusterize( self, distance, minpt, onlyhot=True ) :
        if not self.good :
            return self

        seglabels = self._retrieve_labels( onlyhot )

        if this.verbose :
            logger.log( "Searching for superclusters, distance: {}, minimum points: {}...".format(distance, minpt) )

        self.sprclus = dict()
        self.sprnois = dict()
        for lbl in seglabels :
            if self.segm.centers[lbl][0] < self.segm.minval :
                continue

            #rmean = np.array( [ c.rmean for c in self.clusters[lbl] ] ).sum() / len( self.clusters[lbl] )
                
            #clusters = [ [ c.bary[0], c.bary[1], c.tmean ] for c in self.clusters[lbl] ]
            clusters = [ [ c.bary[0], c.bary[1] ] for c in self.clusters[lbl] ]
            #distance *= rmean

            C = np.array( self.clusters[lbl] )
            S = np.array( clusters )

            if this.verbose :
                logger.log( "Fitting superclusters for segment {}, distance rescaled to {:.1f} ({:.1f}°C)".format(lbl, distance, self.segm.centers[lbl][0]), end=' ', flush=True )

            labels = sklearn.cluster.DBSCAN( eps=distance, min_samples=minpt ).fit_predict( S )

            unique_labels = set( labels )
            clusnum = len( unique_labels )
            noisenum = 0
            if -1 in labels :
                clusnum -= 1
                noisenum = list( labels ).count( -1 )

            if this.verbose :
                logger.log( "Done! Found {} superclusters over {} clusters, {} noise...".format(clusnum, len(self.clusters[lbl]), noisenum) )

            self.sprclus[lbl] = []
            self.sprnois[lbl] = []

            for sn in unique_labels :
                if sn == -1 :
                    if this.verbose :
                        logger.log( "Skipping noise supercluster" )

                    indices = [ i[0] for i in enumerate(labels) if i[1] == -1 ]
                    self.sprnois[lbl] = [ [ pt[0], pt[1], pt[2] ] for pt in S[indices] ]
                else :
                    indices = [ i[0] for i in enumerate(labels) if i[1] == sn ]

                    c = Cluster( lbl, sn )
                    c.set_clusters( C[indices] )

                    self.sprclus[lbl].append( c )

            self.sprclus[lbl].sort( key=lambda c: c.size(), reverse=True )
            if this.verbose :
                for c in self.sprclus[lbl] :
                    logger.log( "Supercluster {:02d}:{:03d} => T = {:4.1f} ± {:2.1f}, center = ({:.1f}, {:.1f}), radius = {}, size: {}".
                                format(lbl, c.cluster, c.tmean, c.trms, c.bary[0], c.bary[1], c.radius, c.size()) )

        return self

    def detect_borders( self, iters, minpts ) :
        if self.sprclus is None :
            clusters = self.clusters
            name = ""
        else :
            clusters = self.sprclus
            name = "super"

        self.borders = { lbl: [] for lbl in clusters }
        image = self._base_image()

        for lbl in clusters :
            for c in clusters[lbl] :
                if this.verbose :
                    logger.log( "Detecting borders of {}cluster {:02d}:{:03d} ({} pixels)".format(name, lbl, c.cluster, c.size()) )

                base = np.array( [ -1000 ] * self.segm.size ).reshape( self.segm.shape )
                c.fill_image( base, value=1000 )

                oldpts = np.array( [] )
                cbor = None
                for n in range(iters) :
                    if this.verbose :
                        logger.log( "-> Iteraction {}/{} ({})...".format(n+1, iters, c.size()), end=' ' )

                    border = skimage.filters.sobel( base )
                    pts = np.argwhere( border > 0 )
                    for p in pts :
                        base[p[0], p[1]] = -1000

                    oldpts = np.append( oldpts, pts ).reshape( (-1, 2) )
                    points = [ list(p) for p in pts ]

                    if this.verbose :
                        logger.log( "Removing border ({} points, {} total, border: {})...".format(len(points), len(oldpts), 0 if cbor is None else cbor.size()), end=' ' )

                    cbor = c.remove_border( points, cbor, image=image )
                    if this.verbose :
                        logger.log( "Done ({} - {})!".format(c.size(), 0 if cbor is None else cbor.size()) )

                    if c.size() < minpts :
                        if this.verbose :
                            logger.log( "!! Iteraction {}/{} interrupted as {}cluster size is now too small ({}/{})".format(n+1, iters, name, c.size(), minpts) )

                        if cbor is not None :
                            cbor.append_cluster_data( c, False, image=image )

                        break

                c.recalculate()
                if cbor is not None :
                    if this.verbose :
                        logger.log( "Border of {}cluster {:02d}:{:03d} has {} pixels.".format(name, lbl, cbor.cluster, cbor.size()) )

                    self.borders[lbl].append( cbor.recalculate() )

            if this.verbose :
                logger.log( "Removing clusters with less than {} pts from segment {}...".format(minpts, lbl), end=' ' )
                logger.log( "Before: {}".format(len(clusters[lbl])), end=' ' )

            clusters[lbl] = [ c for c in clusters[lbl] if c.size() >= minpts ]

            if this.verbose :
                logger.log( "After: {}".format(len(clusters[lbl])) )

        return self

    def noise_grow( self, distance, mintemp=float('NaN'), onlyhot=True ) :
        if not self.good :
            return self

        seglabels = self._retrieve_labels( onlyhot )
        
        if this.verbose :
            logger.log( "Recovering noise pixels, maximum distance: {}".format(distance) )

        if self.sprclus is None :
            clusters = self.clusters
        else :
            clusters = self.sprclus

        noises = self.noises
            
        for lbl in seglabels :
            if self.segm.centers[lbl][0] < self.segm.minval :
                if this.verbose :
                    logger.log( "Skipping recovering of segment {}: NODATA".format(lbl) )

                continue

            if this.verbose :
                logger.log( "Recovering noise of segment {}: {:.1f}°C, {} pixels...".format(lbl, self.segm.centers[lbl][0], len(noises[lbl])) )

            number = len( noises[lbl] )
            labels = [ -1 ] * number
            count = loop = 1
            while count != 0 :
                count = total = 0

                if this.verbose :
                    logger.log( "Loop {}, examining noise...".format(loop), end='', flush=True )
                    
                for i in range(number) :
                    if labels[i] != -1 :
                        continue

                    n = noises[lbl][i]

                    total += 1                
                    dist = ni = None

                    for ci in range(len(clusters[lbl])) :
                        d = clusters[lbl][ci].distance( n[1], n[0], n[2] )

                        if (d is not None) and ((d < distance) or (not np.isnan(mintemp) and (n[2] >= mintemp))) :
                            if (dist is None) or (d < dist) :
                                dist = d
                                ni = ci

                    if ni is not None :
                        c = clusters[lbl][ni]
                        c.grow( n )

                        labels[i] = c.cluster
                        count += 1
#                    elif this.verbose :
#                        if np.isnan(mintemp) :
#                            logger.log( "Pixel {} not recovered!".format(n) )
#                        else :
#                            logger.log( "Pixel {} not recovered {:.1f}".format(n, mintemp) )

                if count > 0 :
                    if this.verbose :
                        logger.log( " Done! Recovered {} pixels over a total of {}!".format(count, total) )
                        logger.log( "Loop {}: growing clusters...".format(loop), end='', flush=True )

                    for c in clusters[lbl] :
                        c.grow()

                    if this.verbose :
                        logger.log( " Done!" )

                    loop += 1
                elif this.verbose :
                    logger.log( " Done! No pixel recovered!" )

            noise = [ l for l in labels if l == -1 ]
            if this.verbose :
                logger.log( "Unrecovered noise pixels: {} over {}".format(len(noise), len(labels)) )

            if self.sprclus is None :
                self.clusters[lbl].sort( key=lambda c: c.size(), reverse=True )
                
                if this.verbose :
                    for c in self.clusters[lbl] :
                        logger.log( "Grown cluster {}:{} => T = {:4.1f} ± {:2.1f}, center = ({:.1f}, {:.1f}), radius = {}, size: {}".
                                    format(lbl, c.cluster, c.tmean, c.trms, c.bary[0], c.bary[1], c.radius, c.size()) )
            else :
                self.sprclus[lbl].sort( key=lambda c: c.size(), reverse=True )
                
                if this.verbose :
                    for c in self.sprclus[lbl] :
                        logger.log( "Grown supercluster {}:{} => T = {:4.1f} ± {:2.1f}, center = ({:.1f}, {:.1f}), radius = {}, size: {}".
                                    format(lbl, c.cluster, c.tmean, c.trms, c.bary[0], c.bary[1], c.radius, c.size()) )

        return self

    def create_flux_clusters( self, fluxer, area, onlyhot=True ) :
        if not self.good :
            return self

        seglabels = self._retrieve_labels( onlyhot )

        if this.verbose :
            logger.log( "Calculating cluster heat power..." )

        self.powerclus = dict()
        for lbl in seglabels :
            if self.sprclus is None :
                self.powerclus[lbl] = fluxer.cluster_flux( self.clusters[lbl], area )
            else :
                self.powerclus[lbl] = fluxer.cluster_flux( self.sprclus[lbl], area )

            self.powerclus[lbl].sort( key = lambda c: c.avg_power().value, reverse=True )

            if this.verbose :
                logger.log( "Heat power of segment {}...".format(lbl) )
                
                for c in self.powerclus[lbl] :
                    orig = c.cluster
                    logger.log( "-> Cluster {}:{}, average temperature: {:.1f} ± {:.1f} °C, total power: {} W over {:.1f} m² (average: {} W/m²)".
                                format(orig.segment, orig.cluster, orig.tmean, orig.trms, c.power, c.area(), c.avg_power()) )

        return self

    def _base_image( self, fill=float('NaN') ) :
        if self.segm.shadow is None :
            transparent = np.array( self.segm.data )
        else :
            transparent = np.array( self.segm.shadow )
            
        transparent[np.isnan(transparent)] = self.segm.nanval
        transparent[transparent < self.segm.minval] = fill
        transparent = transparent.reshape( self.segm.shape )

        return transparent
    
    def draw_power( self, powerfile=None, avgfile=None ) :
        if not this.draw and powerfile is None and avgfile is None :
            return self

        if this.fluxmap is None :
            this.fluxmap = create_fluxmap()
            
        unique_labels = [ lbl for lbl in self.segm.unique if (lbl in self.powerclus) and (len(self.powerclus[lbl]) != 0) ]
        number = len( unique_labels )

        transparent = self._base_image()

        if number > 1 :
            figshape = best_geometry( number )
            figsize = tuple( e1 * e2 for e1, e2 in zip(figshape, self.segm.figsize) )
        else :
            figsize = self.segm.figsize

        figure = plt.figure( figsize=figsize )

        for segi in range(number) :
            lbl = unique_labels[segi]
            if self.segm.centers[lbl][0] < self.segm.minval :
                continue

            if number > 1 :
                ax = figure.add_subplot( figshape[0], figshape[1], segi + 1 )
                ax.set_title( "Image segment {}: {:.1f}°C".format(lbl, self.segm.centers[lbl][0]) )

                plt.imshow( transparent, vmin=self.segm.minval, vmax=self.segm.maxval, cmap=plt.cm.gist_gray, extent=self.segm.extent )
                if self.segm.extent is not None :
                    utm_axis( ax )
            else :
                if not this.large :
                    plt.title( "Image segment {}: {:.1f}°C".format(lbl, self.segm.centers[lbl][0]) )

                plt.imshow( transparent, vmin=self.segm.minval, vmax=self.segm.maxval, cmap=plt.cm.gist_gray, extent=self.segm.extent )
                if self.segm.extent is not None :
                    utm_axis( plt.gca() )

            clsimage = np.array( [ float('NaN') ] * self.segm.size ).reshape( self.segm.shape )

            for c in self.powerclus[lbl] :
                c.fill_image( clsimage, mean=False )

            plt.imshow( clsimage, cmap=this.fluxmap, extent=self.segm.extent )
            cbar = plt.colorbar()
            if this.large :
                cbar.set_label( "Total emitted power (W)", rotation=270, labelpad=15, size=18 )
                cbar.ax.tick_params( labelsize=18 )
            else :
                cbar.set_label( "Total emitted power (W)", rotation=270, labelpad=15 )

        if not this.large :
            if this.title is None :
                if number > 1 :
                    figure.suptitle( "Cluster total power for each segment" )
                else :
                    figure.suptitlt( "Cluster total power" )
            elif number > 1 :
                figure.suptitle( "{}\nCluster total power for each segment".format(this.title) )
            else :
                figure.suptitle( "{}\nCluster total power".format(this.title) )

        draw_figure( figure, powerfile )

        figure = plt.figure( figsize=figsize )

        for segi in range(number) :
            lbl = unique_labels[segi]
            if self.segm.centers[lbl][0] < self.segm.minval :
                continue

            if number > 1 :
                ax = figure.add_subplot( figshape[0], figshape[1], segi + 1 )
                ax.set_title( "Image segment {}: {:.1f}°C".format(lbl, self.segm.centers[lbl][0]) )

                plt.imshow( transparent, vmin=self.segm.minval, vmax=self.segm.maxval, cmap=plt.cm.gist_gray, extent=self.segm.extent )
                if self.segm.extent is not None :
                    utm_axis( ax )
            else :
                if not this.large :
                    plt.title( "Image segment {}: {:.1f}°C".format(lbl, self.segm.centers[lbl][0]) )

                plt.imshow( transparent, vmin=self.segm.minval, vmax=self.segm.maxval, cmap=plt.cm.gist_gray, extent=self.segm.extent )
                if self.segm.extent is not None :
                    utm_axis( plt.gca() )

            clsimage = np.array( [ float('NaN') ] * self.segm.size ).reshape( self.segm.shape )

            for c in self.powerclus[lbl] :
                c.fill_image( clsimage, mean=True )

            plt.imshow( clsimage, cmap=this.fluxmap, extent=self.segm.extent )
            cbar = plt.colorbar()
            if this.large :
                cbar.set_label( "Average emitter power (W/m$^2$)", rotation=270, labelpad=15, size=18 )
                cbar.ax.tick_params( labelsize=18 )
            else :
                cbar.set_label( "Average emitter power (W/m$^2$)", rotation=270, labelpad=15 )

        if not this.large :
            if this.title is None :
                if number > 1 :
                    figure.suptitle( "Cluster average power for each segment" )
                else :
                    figure.suptitlt( "Cluster average power" )
            elif number > 1 :
                figure.suptitle( "{}\nCluster average power for each segment".format(this.title) )
            else :
                figure.suptitle( "{}\nCluster average power".format(this.title) )

        draw_figure( figure, avgfile )

        return self

    def draw_clusters( self, grid, shapefile=None, centroidfile=None, borders=False ) :
        if not this.draw and (shapefile is None) and (centroidfile is None) :
            if this.verbose :
                logger.log( "Skipping cluster drawing/imageing..." )
                
            return self

        if self.sprclus is not None :
            clusters = self.sprclus
            what = "Superclusters"
        else :
            clusters = self.clusters
            what = "Clusters"

        drawer = ClusterDrawer( self.segm, clusters, self._base_image() )

        if shapefile is not None :
            if borders and (self.borders is not None) :
                logger.log( "Passing borders ({}) while creating file <{}>".format(len(self.borders), shapefile) )
            drawer.set_title( "{} shape".format(what) ).draw_clusters( shapefile, borders=(self.borders if borders else None) )

        if centroidfile is not None :
            drawer.set_title( "{} centroids and mean size".format(what) ).draw_centroids( grid, centroidfile )

        return self

    def draw_borders( self, shapefile=None ) :
        if not this.draw and (shapefile is None) :
            if this.verbose :
                logger.log( "Skipping cluster borders drawing/imageing..." )
                
            return self

        if self.sprclus is None :
            what = "Clusters"
        else :
            what = "Superclusters"

        drawer = ClusterDrawer( self.segm, self.borders, self._base_image() )
        drawer.set_title( "{} border".format(what) ).draw_clusters( shapefile, mean=False, value=self.segm.maxval )

        return self
