#!/usr/bin/python3

import os
import sys

import numpy as np
import matplotlib.pyplot as plt

import heatflux

def read_data( filename ) :
    Points = []
    Thermal = []
    
    with open( filename, 'r' ) as handle :
        for line in handle :
            pt = heatflux.RawPoint( line )

            if pt.good :
                Points.append( pt )

                pt = heatflux.DataPoint( pt )
                Thermal.append( pt.point_heat() )

    return Points, Thermal

def write_file( fig, dirname, name ) :
    if dirname is not None :
        name = os.path.join( dirname, name )

        fig.savefig( name, dpi=300, bbox_inches="tight" )
        print( "Written image file <{}>".format(name) )

    return

Points, Thermal = read_data( sys.argv[1] )
if len(sys.argv) == 3 :
    outdir = sys.argv[2]
else :
    outdir = None

X = np.array( [ [ heatflux.to_celsius(pt.Tair.value), pt.Tair.error, heatflux.to_celsius(pt.Tbg.value), pt.Tbg.error, pt.Tair.value - pt.Tbg.value ]
                for pt in Points if pt.Tair.value > 0 and pt.Tbg.value > 0 and pt.status == 'S' ] )

Y = np.array( [ [ heatflux.to_celsius(p[0].Tbg.value), p[0].Tbg.error, p[1][3].value, p[1][3].error ]
                for p in zip(Points, Thermal) if p[0].Tair.value > 0 and p[0].Tbg.value > 0 and p[0].status == 'S' ] )

mindiff = X[:, 4].min()
maxdiff = X[:, 4].max()
n = int(maxdiff - mindiff) + 1

bars, edges = np.histogram( X[:, 4], n )
hc = (edges[:-1] + edges[1:]) / 2
width = np.diff( edges )

fig = plt.figure()

coeff = np.corrcoef( X[:, 0], X[:, 2] )[0, 1]

#fig.suptitle( "Reprocessed from Marotta et al., 2019 (C = {:.2f})".format(coeff) )
plt.errorbar( X[:, 0], X[:, 2], X[:, 3], X[:, 1], marker='+', ms=4, lw=1, ls='none' )
plt.xlabel( "T$_{air}$ ($^o$C)" )
plt.ylabel( "T$_{bg}$ ($^o$C)" )
plt.grid()
write_file( fig, outdir, "Tbg_vs_Tair.png" )

fig = plt.figure()

#fig.suptitle( "Reprocessed from Marotta et al., 2019" )
plt.bar( hc, bars, align='center', width=width, color='red', edgecolor='black' )
plt.xlabel( "T$_{air}$ - T$_{bg}$ ($^o$C)" )
plt.grid()
write_file( fig, outdir, "DiffTemp.png" )

fig = plt.figure()

coeff = np.corrcoef( Y[:, 0], Y[:, 2] )[0, 1]

#fig.suptitle( "Reprocessed from Marotta et al., 2019 (C = {:.2f})".format(coeff) )
plt.errorbar( Y[:, 0], Y[:, 2], Y[:, 3], Y[:, 1], marker='+', ms=4, lw=1, ls='none' )
plt.xlabel( "T$_{bg}$ ($^o$C)" )
plt.ylabel( "$\Phi$ (W/m$^2$)" )
plt.grid()
write_file( fig, outdir, "Phi_vs_Tbg.png" )

plt.show()
