#!/usr/bin/python3
#  -*- coding: utf-8 -*-

import sys

paths = [ p for p in sys.path if 'local' not in p ]
sys.path = paths

import os
import pprint
import re

import datetime
import matplotlib.pyplot as plt
import numpy as np

import logger
import asciigrid
import clustering
import heatflux
import output
from utils import ValueError

def print_cluster_utm( grid, clusters ) :
    logger.log( "UTM coordinates of clusters:" )
    
    for c in clusters :
        px = grid.pixel2utm( c.to_pixel(), True )
        logger.log( "-> Cluster {}:{}, barycenter ({}, {}) -> {}".format(c.cluster.segment, c.cluster.cluster, c.cluster.bary[0], c.cluster.bary[1], px) )

    return

def check_filename( filename ) :
    #expr = re.compile( "^.+[pP]rodotto([0-9]+)_([0-9]{4})([0-9]{2})([0-9]{2})_[tT]_(.+(?:(?<!\.txt)))(?:\.[a-z]+)+$" ) #_T_([A-Za-z0-9\_]+(?:(?<!\.txt)))\.[a-z]+$", re.DEBUG )
    expr = re.compile( "^.+/([0-9]{4})([0-9]{2})([0-9]{2})_T_([0-9]+)_([^_]+)_([0-9]+)_(.+(?:(?<!\.txt)))(?:\.[a-z]+)+$" )

    result = expr.match( filename )
    if result :
        loc = result.group( 5 )
        spec = result.group( 7 )
        output = { 'year':   int( result.group(1) ),
                   'month':  int( result.group(2) ),
                   'day':    int( result.group(3) ),
                   'flight': int( result.group(4) ),
                   'height': float( result.group(6) ),
                   'loc':    loc, 'spec': spec,
                   'niceloc': "{} - {}".format(loc, spec.replace('_', ' ') )
                  }

    else :
        logger.log( "File <{}> has wrong name format, skipping...".format(filename) )
        output = None

    return output

def create_imagenames( params, parsed ) :
    if params['imgdir'] is None :
        images = {
            'imgdir': None,
            'mapfile': None, 'bgmap': None,
            'segments': None, 'histogram': None,
            'fluxsegs': None, 'fluxhisto': None,
            'shapes': None, 'centroids': None,
            'sprshapes': None, 'sprcents': None, 'sprbordrs': None,
            'thr_segmn': None, 'thr_histo': None,
            'flux': None,
            'power': None, 'avgpower': None
        }

        outdir = None
    else :
        year = parsed['year']
        month = parsed['month']
        day = parsed['day']
        loc = parsed['loc']
        spec = parsed['spec']

        if params['mkdir'] :
            imgdir = os.path.join( params['imgdir'], "{:04d}{:02d}{:02d}_Product{:02d}".format(year, month, day, parsed['prod']) )
            outdir = os.path.join( params['outdir'], "{:04d}{:02d}{:02d}_Product{:02d}".format(year, month, day, parsed['prod']) )
        else :
            imgdir = params['imgdir']
            outdir = params['outdir']
        
        images = {
            'imgdir':     imgdir,
            'mapfile':    os.path.join( imgdir, "{:04d}{:02d}{:02d}_01.01_{}_{}_ThermalMap.png".format(year, month, day, loc, spec) ),
            'enhanced':   os.path.join( imgdir, "{:04d}{:02d}{:02d}_01.02_{}_{}_EnhancedMap.png".format(year, month, day, loc, spec) ),
            'bgmap':      os.path.join( imgdir, "{:04d}{:02d}{:02d}_02_{}_{}_ThermalMap_BG.png".format(year, month, day, loc, spec) ),
            'segments':   os.path.join( imgdir, "{:04d}{:02d}{:02d}_03.{{:02d}}_{}_{}_Segments_{}.png".format(year, month, day, loc, spec, params['segments']) ),
            'histogram':  os.path.join( imgdir, "{:04d}{:02d}{:02d}_04.{{:02d}}_{}_{}_Histogram_{}.png".format(year, month, day, loc, spec, params['segments']) ),
            'flux':       os.path.join( imgdir, "{:04d}{:02d}{:02d}_05.{{:02d}}_{}_{}_HeatFlux.png".format(year, month, day, loc, spec) ),
            'fluxsegs':   os.path.join( imgdir, "{:04d}{:02d}{:02d}_06_{}_{}_Segments_Flux.png".format(year, month, day, loc, spec) ),
            'fluxhisto':  os.path.join( imgdir, "{:04d}{:02d}{:02d}_07_{}_{}_Histogram_Flux.png".format(year, month, day, loc, spec) ),
            'shapes':     os.path.join( imgdir, "{:04d}{:02d}{:02d}_08_{}_{}_Shapes_{}.png".format(year, month, day, loc, spec, params['segments']) ),
            'centroids':  os.path.join( imgdir, "{:04d}{:02d}{:02d}_09_{}_{}_Centroids_{}.png".format(year, month, day, loc, spec, params['segments']) ),
            'thr_segmn':  os.path.join( imgdir, "{:04d}{:02d}{:02d}_10_{}_{}_Segments_Thr_{{:.1f}}.png".format(year, month, day, loc, spec) ),
            'thr_histo':  os.path.join( imgdir, "{:04d}{:02d}{:02d}_11_{}_{}_Histogram_Thr_{{:.1f}}.png".format(year, month, day, loc, spec) ),
            'sprshapes':  os.path.join( imgdir, "{:04d}{:02d}{:02d}_12_{}_{}_Hot_SC_Shapes_{}.png".format(year, month, day, loc, spec, params['segments']) ),
            'sprcentrs':  os.path.join( imgdir, "{:04d}{:02d}{:02d}_13_{}_{}_Hot_SC_Centroids_{}.png".format(year, month, day, loc, spec, params['segments']) ),
            'sprbordrs':  os.path.join( imgdir, "{:04d}{:02d}{:02d}_12_{}_{}_Hot_SC_Borders_{}.png".format(year, month, day, loc, spec, params['segments']) ),
            'power':      os.path.join( imgdir, "{:04d}{:02d}{:02d}_14_{}_{}_Total_Power_{}.png".format(year, month, day, loc, spec, params['segments']) ),
            'avgpower':   os.path.join( imgdir, "{:04d}{:02d}{:02d}_15_{}_{}_Average_Power_{}.png".format(year, month, day, loc, spec, params['segments']) ),
        }

    return images, outdir

def create_directories( imgdir, outdir, verbose=False ) :
    if imgdir is not None :
        if verbose :
            logger.log( "Creating image directory <{}>...".format(imgdir), end=' ', flush=True )

        os.makedirs( imgdir, exist_ok=True )

        if verbose :
            logger.log( "Done!" )

    if outdir is not None :
        if verbose :
            logger.log( "Creating output directory <{}>...".format(outdir), end=' ', flush=True )

        os.makedirs( outdir, exist_ok=True )

        if verbose :
            logger.log( "Done!" )

    return

def create_images_title( parsed ) :
    title = "{} - {:04d}/{:02d}/{:02d}".format( parsed['niceloc'], parsed['year'], parsed['month'], parsed['day'] )
    clustering.set_title( title )

    return title

def read_gridfile( filename, params ) :
    if params['verbose'] :
        logger.log( "Reading file <{}>...".format(filename), end=' ', flush=True )

    clustering.set_large( params['large'] )
    asciigrid.AsciiGrid.set_verbose( params['verbose'] )
    asciigrid.AsciiGrid.set_large( params['large'] )
    grid = asciigrid.AsciiGrid( filename, True )

    if grid.is_good() and params['verbose'] :
        logger.log( "Success!" )
        logger.log( "Retrieving image parameters..." )
        logger.log( "-> Image size ({} x {}) <=> ({:.1f} x {:.1f}) m, data pixels: {}/{}, used surface: {:.1f} m², mean T: {}°C".
                    format(grid.columns, grid.rows, grid.width(), grid.height(), grid.countgood, grid.size(), grid.used_surface(), grid.average()) )
    elif params['verbose'] :
        logger.log( "Failed!" )

    return grid

def fluxed_segmentation( grid, flux_mask, params, images, oldseg ) :
    tempthresh = 12
    percthresh = 20

    hotidx  = oldseg.sortlabs[-1][0]
    hotsize = oldseg.sizes[hotidx]
    hottemp = oldseg.sortlabs[-1][1]
    oldperc = 100. * hotsize / grid.countgood
    
    if params['verbose'] :
        logger.log( "K-means based hot segment occupies {:.1f}% of the used image...".format(oldperc) )
        logger.log( "Trying to obtain a better segmentation using good flux zones..." )
        
    newseg = clustering.Segmenter( grid.data, grid.minimum, grid.maximum, grid.extent )
    newseg.apply_mask( flux_mask )

    newidx  = newseg.sortlabs[-1][0]
    newsize = newseg.sizes[newidx]
    newtemp = newseg.sortlabs[-1][1]
    newperc = 100. * newsize / grid.countgood

    if params['verbose'] :
        logger.log( "Flux-based hot segment occupies {:.1f}% of the used image...".format(newperc) )
        logger.log( "Hottemp {:.1f}, Newtemp {:.1f} ({:.1f})".format(hottemp, newtemp, hottemp - newtemp) )

    if ((hottemp - newtemp) > tempthresh) or (newperc > percthresh) :  # wan and
        if params['verbose'] :
            logger.log( "Flux-based hot segment temperature is too low or area is too big!" )
            logger.log( "Reverting to the use of k-means segments..." )

        seg = oldseg
    else :
        if params['verbose'] :
            logger.log( "Flux-based hot segment temperature is high enough and its area is small!" )
            logger.log( "Abandoning k-means segment..." )

        seg = newseg
        seg.use_size( grid.figsize ).draw_segments( images['fluxsegs'], "Flux based segmentation" ).draw_histogram( images['fluxhisto'], "Flux based segmentation" )

    return seg

def calculate_flux( grid, Tbg, params, fluxer=None, mask=None ) :
    Tair = ValueError( params['tair'], 0.1 )
    rHum = ValueError( params['hum'], 1 )

    logger.nextpass()
    if params['verbose'] :
        fluxtype = "radiative" if params['radiative'] else "complete"

        if fluxer is None :
            logger.log( "Background temperature is {}°C".format(Tbg) )
            logger.log( "Calculating {} heat flux over full image (Tair = {}°C, rH = {}%)...".format(fluxtype, Tair, rHum), end=' ', flush=True )
        else :
            logger.log( "Recalculating {} heat flux only on masked pixels (Tair = {}°C, rH = {}%)...".format(fluxtype, Tair, rHum), end=' ', flush=True )

    if fluxer is None :
        fluxer = heatflux.ThermalImage( grid.data, params['vuerror'] )
    else :
        fluxer.update_image( grid.data )

    fluxer.calculate_flux( Tbg, Tair, rHum, mask )
    if params['verbose'] :
        logger.log( "Done!" )
        
        power = fluxer.total_power( grid.cellarea() )
        flux  = fluxer.total_flux( grid.cellarea() )

        logger.log( "Image total power release: {}W ({:.1f} ± {:.1f}), total mean flux: {}W/m²".format(power, power.value, power.error, flux) )

    if params['verbose'] and (mask is None) :
        logger.nextpass()
        logger.log( "Calculating reduced power and flux over image..." )

    flux_mask, ratio = fluxer.reduced_mask( grid.cellarea(), params['sigma'] )
    if flux_mask.any() :
        if params['verbose'] :
            power = fluxer.reduced_power( grid.cellarea(), params['sigma'] )
            flux  = fluxer.reduced_flux( grid.cellarea(), params['sigma'] )

            logger.log( "Image reduced power release (at {} sigma): {}W ({:.1f} ± {:.1f}), reduced mean flux: {}W/m^2".format(params['sigma'], power, power.value, power.error, flux) )
    else :
        if params['verbose'] :
            logger.log( "Irreducible image, aborting (ratio extremes [{:.1e}, {:.1e}]...".format(ratio.min(), ratio.max()) )

        flux_mask = None

    return flux_mask, (fluxer)

def preventive_segmentation( grid, params, images, cut_delta ) :
    counter = 1
    number = params['segments']
    lowidx = None
    preseed = False

    if (params['tair'] is not None) and (params['hum'] is not None) and params['verbose'] :
        tair = ValueError( params['tair'], 0.1 )
        hum  = ValueError( params['hum'], 1 )

        if params['height'] is None :
            logger.log( "|-> Environmental parameters: Tair = {}°C, rH = {}%".format(tair, hum) )
        else :
            logger.log( "|-> Environmental parameters: Tair = {}°C, rH = {}%, flight quote: {}m".format(tair, hum, params['height']) )
    
    seg = clustering.Segmenter( grid.data, grid.minimum, grid.maximum, grid.extent )
    seg.use_size( grid.figsize )

    while True :
        logger.nextpass()

        seg.segment( number, preseed=preseed, loop=counter )
        seg.draw_segments( images['segments'].format(counter) ).draw_histogram( images['histogram'].format(counter) )

        if params['tair'] is None :
            lowidx, lowseg, lowt = seg.minimum_segment()
            break
        else :
            if lowidx is None :
                lowidx, lowseg, lowt = seg.minimum_segment()
            else :
                lowidx += 1
                lowseg, lowt = seg.segment_by_index( lowidx )

            if (params['tair'] - lowt) <= cut_delta :
                if params['verbose'] :
                    logger.log( "Low segment temperature is within {:.1f}°C from Tair, segmentation terminated...".format(cut_delta) )

                break
            else :
                if params['verbose'] :
                    logger.log( "Difference between air temperature and low temperature is over {:.1f}°C, adding a segment...".format(cut_delta) )

                number += 1
                counter += 1
                preseed = True
    
    return seg, (lowseg, lowt)

def initialize_analysis( filename, params, prod ) :
    parsed = check_filename( filename )
    if parsed is None :
        return None, None
    elif (prod is not None) and (prod != 0) :
        parsed['prod'] = prod

    cut_delta = 5
    flightdate = datetime.date( parsed['year'], parsed['month'], parsed['day'] )
    title = create_images_title( parsed )
    images, outdir = create_imagenames( params, parsed )

    grid = read_gridfile( filename, params )
    if grid.is_good() :
        output.set_verbose( params['verbose'] )
        clustering.set_verbose( params['verbose'] )
        clustering.enable_draw( params['draw'] )

        if params['mkdir'] :
            create_directories( images['imgdir'], outdir, params['verbose'] )

        out = output.Output( outdir, filename, flightdate )

    return grid, (parsed, images, outdir, title, out, cut_delta)

def extract_background( grid, seg, lowseg, params, images ) :
    if params['bgpoint'] is None :
        if params['verbose'] :
            logger.log( "Calculating background temperature on \"low\" segment {}...".format(lowseg) )

        clus = clustering.Clusterer( seg )
        if not clus.good :
            logger.log( "Error initializing clustering. Abort!" )
            return

        clus.clusterize( params['radius'], params['minsize'], lowseg )

        Tbg = clus.extract_background( segment=lowseg )
        if Tbg is None :
            logger.log( "Cannot extract background temperature. Aborting!" )
            return
    else :
        if params['verbose'] :
            logger.log( "Requested calculation of background temperature at point ({:.2f}, {:.2f}) for {:.2f}m".format(*params['bgpoint']) )

        bgmask = grid.circle_mask( params['bgpoint'][:2], params['bgpoint'][2] )
        if bgmask.any() :
            Tbg = grid.average( bgmask, True )
            grid.draw( images['bgmap'], params['draw'], hmask=bgmask )
        else :
            Tbg = None

            if params['verbose'] :
                logger.log( "Requested background area is not contained in current image. Aborting..." )

    return Tbg

def analyze_flight( filename, params, prod ) :
    if params['verbose'] :
        logger.log( "Processing file <{}>...".format(filename) )

    logger.startpass()

    grid, others = initialize_analysis( filename, params, prod )
    if (grid is None) or not grid.is_good() :
        return

    (parsed, images, outdir, title, out, cut_delta) = others

    average = grid.average()
    grid.draw( images['mapfile'], params['draw'] )

    seg, (lowseg, lowt) = preventive_segmentation( grid, params, images, cut_delta )
    if (params['tair'] is None) or (params['hum'] is None) :
        return

    logger.nextpass()
    if params['skyair'] :
        if params['verbose'] :
            logger.log( "Using air temperature as the sky temperature to calculate radiative flux" )
            logger.log( "Simulating Tbg as Tair (it is not used)" )

        heatflux.use_tair_as_tsky()
        Tbg = ValueError( params['tair'], 0.1 )
    elif params['radiative'] :
        if params['verbose'] :
            logger.log( "Using only radiative part to calculate heat flux. Air temperature is used as Tsky" )
            logger.log( "Simulating Tbg as Tair (it is not used)" )

        heatflux.only_calculate_radiative()
        Tbg = ValueError( params['tair'], 0.1 )
    else :
        Tbg = extract_background( grid, seg, lowseg, params, images )

        if Tbg is None :
            return

    flux_mask, (fluxer) = calculate_flux( grid, Tbg, params )
    
    if flux_mask is None :
        if fluxer is not None :
            seg.draw_fluximage( fluxer, images['flux'].format(1), fluxmap=plt.cm.gnuplot2 )
            
        return

    logger.nextpass()
    seg = fluxed_segmentation( grid, flux_mask, params, images, seg )
    seg.draw_fluximage( fluxer, images['flux'].format(1), plt.cm.gnuplot2 )

    # Move height correction after fluxed segmentation
    if params['loss'] > 0 :
        logger.nextpass()

        enhance = clustering.Enhancer( seg )
        if enhance.good :
            enhance.enhance( params['height'], params['loss'], clustering.Enhancer.ENHANCE_HIGH )
            enhance.draw_enhanced( images['enhanced'] )

            grid.update_image( enhance.data, enhance.minval, enhance.maxval )

            calculate_flux( grid, Tbg, params, fluxer, flux_mask )
            seg.draw_fluximage( fluxer, images['flux'].format(2), plt.cm.gnuplot2, "Corrected pixels" )

    clus = clustering.Clusterer( seg )
    if not clus.good :
        logger.clearpass()
        logger.log( "Error initializing reclustering. Abort!" )

        return

    logger.nextpass()
    clus.clusterize( params['radius'], params['minsize'], clustering.Clusterer.CLUSTERIZE_HIGH )
    clus.draw_clusters( grid, images['shapes'], images['centroids'] )

    if params['warp'] and params['verbose'] :
        logger.log( "Warping is not necessary after sigma based clustering." )

    if params['norecluster'] :
        logger.log( "Skipping reclustering." )
    else :
        logger.nextpass()
        clus.reclusterize( 2. * params['radius'] / grid.cellsize, 1 )

    if params['borders'] > 0 :
        logger.nextpass()
        if params['verbose'] :
            logger.log( "Detecting superclusters borders and removing them." )

        clus.detect_borders( params['borders'], params['minsize'] ).draw_borders( images['sprbordrs'] )

    elif params['nogrow'] :
        if params['verbose'] :
            logger.log( "Skipping noise pixel recover." )

    else :
        logger.nextpass()
        clus.noise_grow( params['noisedist'], grid.minimum )

    if not params['norecluster'] :
        clus.draw_clusters( grid, images['sprshapes'], images['sprcentrs'] )

    logger.nextpass()
    clus.create_flux_clusters( fluxer, grid.cellarea() ).draw_power( images['power'], images['avgpower'] )

    if params['verbose'] :
        logger.nextpass()

    out.set_data( grid, clus, fluxer )
    if 'prod' in parsed :
        out.set_product( parsed['prod'] )
    out.write_gridfiles().write_csvfiles().write_datfiles()

    return

def print_parameters( filename, params, command ) :
    if command is not None :
        logger.log( "Current command:" )
        logger.log( command )

    logger.log( "Analysis parameters:" )
    logger.log( "- Output file directory: {}".format(params['outdir']) )
    logger.log( "- Image file directory: {}".format(params['imgdir']) )
    if ('directory' in params) and (params['directory'] is not None) :
        logger.log( "- Input files directory: {}".format(params['directory']) )
    if params['mkdir'] :
        logger.log( "- Subdirectories creation enabled" )
    else :
        logger.log( "- Subdirectories creation disabled" )
    if params['draw'] :
        logger.log( "- Real time image display enabled" )
    else :
        logger.log( "- Real time image display disabled" )
    if ('noimages' in params) and params['noimages'] :
        logger.log( "- Image files creation disabled" )
    else :
        logger.log( "- Image files creation enabled" )

    if ('large' in params) and params['large'] :
        logger.log( "- Drawing article-quality images" )

    logger.log( "---------\n- Minimum number of segments: {}".format(params['segments']) )
    logger.log( "- Minimum cluster size: {}".format(params['minsize']) )
    logger.log( "- Maximum pixel distance for cluster: {}".format(params['radius']) )
    logger.log( "- Maximum pixel distance for noise recover: {}".format(params['noisedist']) )
    logger.log( "- Temperature error on thermal camera: {}°C".format(params['vuerror']) )
    logger.log( "- Sigma threshold for heat flux: {}".format(params['sigma']) )
    if (params['loss'] is None) or (params['loss'] <= 0) :
        logger.log( "- No temperature loss per meter set: recovery will not be performed." )
    else :
        logger.log( "- Temperature loss per meter: {}%/m.".format(params['loss']) )
    if (params['maxheight'] is None) or (params['maxheight'] <= 0) :
        logger.log( "- Maximum height not set." )
    else :
        logger.log( "- Maximum accepted height set to {}".format(params['maxheight']) )
    if params['bgpoint'] is None :
        logger.log( "- Background point not used: using low temperature segment to extract Tbg" )
    else :
        logger.log( "- Background temperature retrieved at ({} {} @{})".format(*params['bgpoint']) )
    logger.log( "---------" )
    if params['radiative'] :
        logger.log( "- Calculating only the radiative part of the flux" )
        logger.log( "- Using Tair as TSky" )
    else :
        logger.log( "- Calculating all components of heat flux" )

        if params['skyair'] :
            logger.log( "- Using Tair as TSky" )
        else :
            logger.log( "- Calculating TSky from zero flux points" )
    if params['borders'] > 0 :
        logger.log( "- Maximum number of border removal iterations: {}".format(params['borders']) )
        logger.log( "- Discarding noise pixels" )
    else :
        logger.log( "- Cluster border removal disabled" )

        if params['nogrow'] :
            logger.log( "- Discarding noise pixels" )
        else :
            logger.log( "- Actively recovering noise pixel after clusterization" )
    logger.log( "---------" )

    return

def single_flight( filename, params, prod=0, command=None ) :
    filexpr = re.compile( "^(?:.+/)?([^/]+(?:(?<!\.txt)))(?:\.txt){1,2}$" ) # , re.DEBUG )
    result = filexpr.match( filename )

    if result :
        name = result.group( 1 )
        if params['outdir'] is not None :
            logfile = os.path.join( params['outdir'], "{}.log".format(name) )
            handle = logger.openfile( logfile, params['verbose'] )

            print_parameters( filename, params, command )

        analyze_flight( filename, params, prod )
        logger.close( handle )

        if params['draw'] :
            plt.show()
    else :
        logger.log( "Match result:" )
        logger.log( result )

    return

if __name__ == "__main__" :
    if len(sys.argv) != 6 :
        print( "Usage: {} <filename> <air T> <relhum> <height> <product>".format(sys.argv[0]) )
    else :
        outdir = "../dronedata/RESULTS/article"
        
        params = dict(
            segments=3, radius=10, minsize=8,
            loss=0.45, sigma=3, maxheight=75,
            nogrow=True, norecluster=True, mkdir=True, verbose=True,
            imgdir=os.path.join(outdir, "images"), outdir=os.path.join(outdir, "output"),
            noisedist=100, vuerror=5., borders=0,
            draw=False, noimages=False, warp=False, bgpoint=None,
            skyair=False, radiative=False, large=True,
        )

        params['tair']   = float( sys.argv[2] )
        params['hum']    = float( sys.argv[3] )
        params['height'] = float( sys.argv[4] )
        product = int( sys.argv[5] )

        logfile = os.path.join( outdir, "single.log" )
        
        """
        params = dict( pixelerr=2., vuerror=5., sigma=2.,
                       segments=3, radius=10, minsize=8, noisedist=100,
                       bgpoint=None,
                       imgdir="../dronedata/test_img", outdir="../dronedata/test_out",
                       draw=False, verbose=True, warp=False, mkdir=True,
                       tair=float(sys.argv[2]), hum=float(sys.argv[3]),
                       skyair=False, radiative=True, nogrow=False, borders=5 )


        params = {
            'pixelerr': 2., 'vuerror': 5., 'sigma': 2.,
            'segments': 3,
            'radius':  10, 'minsize': 8,
            'noisedist': 100,
            'bgpoint': None,
            'imgdir': None, 'outdir': None,
            'imgdir': "../dronedata/test_img", 'outdir': "../dronedata/test_out",
            'draw': False, 'verbose': True, 'warp': False, 'mkdir': True,
            'tair': float(sys.argv[2]), 'hum': float(sys.argv[3]),
            'skyair': False, 'radiative': False
        }
        """

        single_flight( sys.argv[1], params, product, " ".join(sys.argv) )
