#!/usr/bin/python3
#  -*- coding: utf-8 -*-

import sys
import math

this = sys.modules[__name__]

#
# These are immutable constants
#
AZO = 273.15                  # Absolute Zero
standard_Pair = 1.01325e5     # Standard air pressure
CC1 = 7.1128e-5               # Conducivity constant 1
univ_R = 8.314472             # Universal gas constant
molw_air = 0.029              # Molecular weight of air (kg/mole)
molw_wat = 0.018              # Molecular weight of water (kg/mole)
air_R = (univ_R / molw_air)   # Gas constant of air
wat_R = (univ_R / molw_wat)   # Gas constant ov water vapour
wat_lh = 2.501e6              # Latent heat of liquid->vapour
beta_air = 3.69e-3            # beta constants of air
kvisc_air = 1.35e-5           # air viscosity constant
grav = 9.81                   # gravity constant (m/s^2)
TAMB = 29.65                  # standard temperature
sb_constant = 5.67e-8

#
# Nusselt constants [ amplitude, exponent ]
#
NU = [ [ 0.54, 0.25 ], [ 0.15, 1./3. ], [ 0.27, 0.25 ] ]

#
# These are not constants but are univoque for each
# measurement
#
this.Tair = TAMB + AZO #  Air temperature
this.dTair = 0.1       #  - error
this.rHum = 10.        #  Relative humidity
this.drHum = 1.        #  - error
this.Tbg  = TAMB + AZO #  Background temperature
this.dTbg = 1.         #  - error
this.lScale = 10       #  Scale dimension (m)
this.ReEm = 0.98       # relative emissivity
this.hconv_b = None
this.heat_conv_b = None
this.heat_evap_b = None
this.Tsky4 = None
this.dTsky4 = None

def air_temperature( value, error=0.1 ) :
    this.Tair = to_kelvin( value )
    this.dTair = error

    return

def relative_humidity( value, error=1. ) :
    this.rHum = value
    this.drHum = error

    return

def background( value, error=1. ) :
    this.Tbg = to_kelvin( value )
    this.dTbg = error

    return

def scale_length( value ) :
    this.lScale = value

    return

def relative_emissivity( value ) :
    this.ReEm = value

    return

def to_kelvin( C ) :
    return C + AZO

def to_celsius( K ) :
    return K - AZO

def square( num ) :
    return num * num

def cube( num ) :
    return num * num * num

def doublesquare( num ) :
    return num * num * num * num

class ValueError :
    def __init__( self, value, error ) :
        self.value = value
        self.error = error

        return

def nusselt_index( Ts, Tair, rayleigh ) :
    idx = 2

    if Ts > Tair :
        if rayleigh < 1e7 :
            idx = 0
        else :
            idx = 1

    return idx

def mean_temperature( t1, dt1, t2, dt2 ) :
    return ValueError( (t1 + t2) / 2, math.sqrt(square(dt1) + square(dt2)) / 2 )

def air_thermal_conducivity( kelvin ) :
    return 0.023807 + CC1 * to_celsius( kelvin ) # Eq. 2.5 Jacobson, 2005

def vapour_thermal_conducivity( kelvin ) :
    return 0.015606 + 8.3680e-5 * to_celsius( kelvin ) # Eq.2.6 Jacobson, 2005

def air_thermal_conducivity_error( dk ) :
    return CC1 * dk

def air_density( T, P ) :
    return P / (air_R * T)

def air_density_error( T, P ) :
    return T.error * math.abs( P / (air_R * T.value * T.value) )

def air_spec_heat( T ) : # Specific heat of dry air at constant pressure (J/Kg K)
    return 1004.67  # At 298 K

def air_sh_error( T ) :
    return 0

def air_thermal_diffusivity( T, P, Kconv_air ) :
    air_dens = ValueError( air_density(T.value, P), air_density_error(T, P) )
    air_sh = ValueError( air_spec_hear(T), air_sh_error(T) )

    Kdiff_air = Kconv_air.value / (air_dens * air_sh)
    dKdiff_air = math.sqrt( square(Kconv_air.error / (air_dens.value * air_sh.value)) +
                            square(Kconv_air.value * air_dens.error / (air_sh.value * square(air_dens.value))) +
                            square(Kconv_air.value * air_sh.error / (air_dens.value * square(air_sh.value))) )

    return ValueError( Kdiff_air, dKdiff_air )

def convective_heat_coefficient( Ts, dTs, Tair, dTair, scale ) :
    tmean = mean_temperature( Ts, dTs, Tair, dTair )

    Kconv_air  = ValueError( air_thermal_conductivity(tmean.value), air_thermal_conducivity_error(tmean.error) )
    Kdiff_air  = air_thermal_diffusivity( tmean, standard_Pair, Kconv_air )

    ray_const  = grav * beta_air * cube(scale) / (kvisc_air * Kdiff_air.value)
    rayleigh   = ray_const * math.abs(Ts - Tair)
    d_rayleigh = ray_const * math.sqrt( square(dTs) + square(dTair) + square(((Ts - Tair) * Kdiff_air.error) / Kdiff_air.value) )

    nuid = nusselt_index( Ts, Tair, rayleigh )
    nusselt = NU[nuid][0] * math.pow( rayleigh, NU[nuid][1] )
    d_nuss  = d_rayleigh * math.abs(NU[nuid][0] * NU[nuid][1] * math.pow(rayleigh, NU[nuid][1] - 1))

    hconv = nusselt * Kconv_air.value / scale
    d_hconv = math.sqrt( square(Kconv_air.error * nusselt) + square(Kconv_air.value * d_nuss) ) / scale

    return ValueError( hconv, d_hconv )

def saturation_pressure_over_liquid( T ) :
    c1 = 611.22
    c2 = 17.67

    Tstar = to_celsius( T.value ) / ( T.value - TAMB )

    psat   = c1 * math.exp( c2 * Tstar )  # Saturation pressure
    d_psat = T.error * math.abs( c2 * psat * (AZO - TAMB) / square(T.value - TAM) )

    return ValueError( psat, d_psat )

def vapour_diffusion_coefficient( T, P ) :
    c1 = 2.11e-5
    c2 = 1.94

    c3 = c1 * standard_Pair / P
    trid = T.value / AZO

    return ValueError( c3 * math.pow(trid, c2), T.error * math.abs(c3 * c2 * math.pow(trid, c2 - 1) / AZO) )

def partial_pressure( P, H ) :
    return ValueError( P.value * H.value / 100, main.sqrt(square(P.value * H.error) + square(P.error * H.value)) / 100 )

def vapour_density( T, P ) :
    return ValueError( P.value / (wat_R * T.value), math.sqrt(square(p.error) + square(P.value * T.error / T.value)) / (wat_R * T.value) )

def _convective_heat( T, hconv ) :
    qconv   = hconv.value * (T.value - this.Tair)
    d_qconv = math.sqrt( square((T.value - this.Tair) * hconv.error) + square(hconv.value) * (square(T.error) + square(this.dTair)) )

    return ValueError( qconv, d_conv )

def _evaporative_heat( T, hconv, Tair, hum ) :
    tmean = mean_temperature( T.value, T.error, Tair.value, Tair.error )
    psat_soil = saturation_pressure_over_liquid( T )                           # Saturation pressure at the soil
    psat_air  = saturation_pressure_over_liquid( Tair )                        # Saturation pressure in the air
    pwat_air  = partial_pressure( psat_air, hum )                              # Partial water pressure in the air
    rhowat_soil = vapour_density( T, psat_soil )                               # Water vapour density at soil (pwat_soil = psat_soil as we consider hum = 100% there
    rhowat_air  = vapour_density( Tair, pwat_air )                             # Water vapour density in the air

    qeva = d_qeva = 0.0
    if rhowat_soil.value >= rhowat_air.value :
        # There is heat transfer only if there is a positive gradient of water vapour density
        Kconv_air  = ValueError( air_thermal_conductivity(tmean.value), air_thermal_conducivity_error(tmean.error) )
        Kdiff_air = air_thermal_diffusivity( tmean, standard_Pair, Kconv_air )
        Kdiff_vap = vapour_diffusion_coefficient( tmean, standard_Pair )
        lewis = ValueError( Kdiff_air.value / Kdiff_vap.value, math.sqrt(square(Kdiff_air.error) + square(Kdiff_air.value * Kdiff_vap.error / Kdiff_vap.value)) / math.abs(Kdiff_vap.value) )
        rho_air = ValueError( air_density(tmean.value, standard_Pair), air_density_error(tmean, standard_Pair) )
        cp_air  = ValueError( air_spec_heat(tmean), air_sh_error(tmean) )

        h_frac = rho_air.value * cp_air.value * math.pow( lewis.value, 2./3. )
        hvap = hconv.value / h_frac
        d_hvap = math.sqrt( square(hconv.error) + square(hconv.value) * (square(rho_air.error/rho_air.value) + square(cp_air.error/cp_air.value) +
								         4. * square(lewis.error/lewis.value) / 9.) ) / h_frac

        qeva = wat_lh * hvap * (rhowat_soil.value - rhowat_air.value)
        d_qeva = wat_lh * math.sqrt( square(d_hvap * (rhowat_soil.value - rhowat_air.value)) + square(hvap) * (square(rhowat_soil.error) + square(rhowat_air.error)) );

    return ValueError( qeva, d_qeva )

def _evaluate_tsky( Tbg, Tair, rHum, hconv, scale ) :
    frac = this.ReEm * sb_constant

    heat_conv_b = _convective_heat( Tbg, hconv )
    heat_evap_b = _evaporative_hear( Tbg, hconv, Tair, rHum )

    # Here was the error: the '-' is a '+'
    # this.Tsky = doublesquare(this.Tbg) - (heat_conv_b.value + heat_evap_b.value) / frac
    Tsky4  = doublesquare(Tbg.value) + (heat_conv_b.value + heat_evap_b.value) / frac
    dTsky4 = math.sqrt( (4 * cube(Tbg.value) * square(Tbg.error)) + ((square(heat_conv_b.error) + square(heat_evap_b.error)) / square(frac)) )

    return ( heat_conv_b, heat_evap_b, Tsky4, dTsky4 )

def evaluate_tsky() :
    if this.hconv_b is None :
        this.hconv_b = convective_heat_coefficient( this.Tbg, this.dTbg, this.Tair, this.dTair, this.lScale )

    ( this.heat_conv_b, this.heat_evap_b, this.Tsky4, this.dTsky4 ) = _evaluate_tsky( ValueError(this.Tbg, this.dTbg),
                                                                                      ValueError(this.Tair, this.dTair),
                                                                                      ValueError(this.rHum, this.drHum),
                                                                                      this.hconv_b, this.lScale )

    return

def evaluate_hconv_bg() :
    this.hconv_b = convective_heat_coefficient( this.Tbg, this.dTbg, this.Tair, this.dTair, this.lScale )

    return

class Point :
    def __init__( self, x, y, ts, dts ) :
        self.x = x
        self.y = y
        self.Ts  = ValueError( to_kelvin(ts), dts )

        self.hconv = convective_heat_coefficient( self.Ts.value, self.Ts.error, this.Tair, this.dTair, this.lScale )

        if this.hconv_b is None :
            this.hconv_b = convective_heat_coefficient( this.Tbg, this.dTbg, this.Tair, this.dTair, this.lScale )

        return

    def convective_heat( self ) :
        return _convective_heat( self.Ts, self.hconv )
    
    def evaporative_heat( self ) :
        return _evaporative_heat( self.Ts, self.hconv )
        
    def radiative_heat( self ) :
        frac = this.ReEm * sb_constant
        if (this.heat_conv_b is None) or (this.heat_evap_b is None) or (this.Tsky4 is None) or (this.dTsky4 is None) :
            evaluate_tsky()

        qrad = frac * (doublesquare(self.Ts.error) - this.Tsky4)
        d_qrad = frac * math.sqrt( (4 * cube(self.Ts.value) * square(self.Ts.error)) + square(this.dTsky4) )

        return ValueError( qrad, d_qrad )

class DataPoint(Point) :
    def __init__( self, pid, x, y, Ts, dTs, Tbg, dTbg, Tair, dTair=this.dTair ) :
        Point.__init__( x, y, Ts, dTs )

        self.Tbg = ValueError( to_kelvin(Tbg), dTbg )
        self.Tair = ValueError( to_kelvin(Tair), dTair )

        self.hconv_b = convective_heat_coefficient( self.Tbg.value, self.Tbg.error, self.Tair.value, self.Tair.error, this.lScale )
