#!/usr/bin/python3
#  -*- coding: utf-8 -*-

import sys
import re
import math
import datetime

import numpy as np

from utils import ValueError, Pixel

this = sys.modules[__name__]
this.tairassky = False
this.onlyrad = False

#
# These are immutable constants
#
AZO = 273.15                  # Absolute Zero
standard_Pair = 1.01325e5     # Standard air pressure
CC1 = 7.1128e-5               # Conducivity constant 1
univ_R = 8.314472             # Universal gas constant
molw_air = 0.029              # Molecular weight of air (kg/mole)
molw_wat = 0.018              # Molecular weight of water (kg/mole)
air_R = (univ_R / molw_air)   # Gas constant of air
wat_R = (univ_R / molw_wat)   # Gas constant ov water vapour
wat_lh = 2.501e6              # Latent heat of liquid->vapour
beta_air = 3.69e-3            # beta constants of air
kvisc_air = 1.35e-5           # air viscosity constant
grav = 9.81                   # gravity constant (m/s^2)
TAMB = 29.65                  # standard temperature
sb_constant = 5.67e-8

#
# Nusselt constants [ amplitude, exponent ]
#
NU = [ [ 0.54, 0.25 ], [ 0.15, 1./3. ], [ 0.27, 0.25 ] ]

#
# These are not constants but are univoque for each
# measurement
#
this.lScale = 10  # Size scale factor
this.ReEm = 0.98  # Relative emissivity

def use_tair_as_tsky( u=True ) :
    this.tairassky = u

    return

def only_calculate_radiative( u=True ) :
    if u :
        this.tairassky = u

    this.onlyrad = u

    return

def to_kelvin( C ) :
    return C + AZO

def to_celsius( K ) :
    return K - AZO

def camera_error( t, e ) :
    if t > 100. :
        err = float(e) * t / 100.
    else :
        err = float(e)

    return err

def global_scale_length( value ) :
    this.lScale = value

    return

def global_relative_emissivity( value ) :
    this.ReEm = value

    return

def square( num ) :
    if isinstance(num, ValueError) :
        v = square( num.value )
        e = 2 * abs(num.value) * num.error

        result = ValueError( v, e )
    else :
        result = num * num

    return result

def cube( num ) :
    if isinstance(num, ValueError) :
        v = cube( num.value )
        e = 3 * square(num.value) * num.error

        result = ValueError( v, e )
    else :
        result = num * num * num

    return result

def doublesquare( num ) :
    if isinstance(num, ValueError) :
        v = doublesquare( num.value )
        e = 4 * abs(cube(num.value)) * num.error

        result = ValueError( v, e )
    else :
        result = num * num * num * num

    return result

def nusselt_index( Ts, Tair, rayleigh ) :
    idx = 2

    if Ts > Tair :
        if rayleigh < 1e7 :
            idx = 0
        else :
            idx = 1

    return idx

def mean_temperature( t1, dt1, t2, dt2 ) :
    return ValueError( (t1 + t2) / 2, math.sqrt(square(dt1) + square(dt2)) / 2 )

def vapour_thermal_conductivity( kelvin ) :
    return 0.015606 + 8.3680e-5 * to_celsius( kelvin ) # Eq.2.6 Jacobson, 2005

def air_thermal_conductivity( K ) :
    atc   = 0.023807 + CC1 * to_celsius( K.value ) # Eq. 2.5 Jacobson, 2005
    d_atc = CC1 * K.error

    return ValueError( atc, d_atc )

def air_density( T, P ) :
    return ValueError( P / (air_R * T.value), T.error * abs( P / (air_R * T.value * T.value) ) )

def air_spec_heat( T ) : # Specific heat of dry air at constant pressure (J/Kg K)
    return ValueError( 1004.67, 0 ) # At 298K

def air_thermal_diffusivity( T, P, Kconv_air ) :
    air_dens = air_density( T, P )
    air_sh   = air_spec_heat( T )

    Kdiff_air  = Kconv_air.value / (air_dens.value * air_sh.value)
    dKdiff_air = math.sqrt( square(Kconv_air.error / (air_dens.value * air_sh.value)) +
                            square(Kconv_air.value * air_dens.error / (air_sh.value * square(air_dens.value))) +
                            square(Kconv_air.value * air_sh.error / (air_dens.value * square(air_sh.value))) )

    return ValueError( Kdiff_air, dKdiff_air )

def convective_heat_coefficient( Ts, Tair, scale ) :
    tmean = mean_temperature( Ts.value, Ts.error, Tair.value, Tair.error )

    Kconv_air  = air_thermal_conductivity( tmean )
    Kdiff_air  = air_thermal_diffusivity( tmean, standard_Pair, Kconv_air )

    ray_const  = grav * beta_air * cube(scale) / (kvisc_air * Kdiff_air.value)
    rayleigh   = ray_const * abs(Ts.value - Tair.value)
    d_rayleigh = ray_const * math.sqrt( square(Ts.error) + square(Tair.error) + square(((Ts.value - Tair.value) * Kdiff_air.error) / Kdiff_air.value) )

    nuid = nusselt_index( Ts.value, Tair.value, rayleigh )
    nusselt = NU[nuid][0] * math.pow( rayleigh, NU[nuid][1] )

    if rayleigh != 0 :
        d_nuss  = d_rayleigh * abs(NU[nuid][0] * NU[nuid][1] * math.pow(rayleigh, NU[nuid][1] - 1))
    else :
        d_nuss  = d_rayleigh * abs(NU[nuid][0] * NU[nuid][1])

    hconv = nusselt * Kconv_air.value / scale
    d_hconv = math.sqrt( square(Kconv_air.error * nusselt) + square(Kconv_air.value * d_nuss) ) / scale

    return ValueError( hconv, d_hconv )

def saturation_pressure_over_liquid( T ) :
    c1 = 611.22
    c2 = 17.67
        
    #Tstar = np.float128( to_celsius(T.value) / (T.value - TAMB) )
    Tstar = to_celsius( T.value ) / ( T.value - TAMB )

    psat   = c1 * np.exp( c2 * Tstar )  # Saturation pressure
    d_psat = T.error * abs( c2 * psat * (AZO - TAMB) / square(T.value - TAMB) )

    return ValueError( psat, d_psat )

def vapour_diffusion_coefficient( T, P ) :
    c1 = 2.11e-5
    c2 = 1.94

    c3 = c1 * standard_Pair / P
    trid = T.value / AZO

    return ValueError( c3 * math.pow(trid, c2), T.error * abs(c3 * c2 * math.pow(trid, c2 - 1) / AZO) )

def partial_pressure( P, H ) :
    return ValueError( P.value * H.value / 100, math.sqrt(square(P.value * H.error) + square(P.error * H.value)) / 100 )

def vapour_density( T, P ) :
    return ValueError( P.value / (wat_R * T.value), math.sqrt(square(P.error) + square(P.value * T.error / T.value)) / (wat_R * T.value) )

def convectiveHeat( Ts, Tair, hconv ) :
    qconv   = hconv.value * (Ts.value - Tair.value)
    d_qconv = math.sqrt( square((Ts.value - Tair.value) * hconv.error) + square(hconv.value) * (square(Ts.error) + square(Tair.error)) )

    return ValueError( qconv, d_qconv )

def evaporativeHeat( Ts, Tair, rHum, hconv ) :
    tmean = mean_temperature( Ts.value, Ts.error, Tair.value, Tair.error )
    psat_soil = saturation_pressure_over_liquid( Ts )       # Saturation pressure at soil
    psat_air  = saturation_pressure_over_liquid( Tair )     # Saturation pressure in the air

    pwat_air = partial_pressure( psat_air, rHum )           # Partial water pressure in the air

    rhowat_soil = vapour_density( Ts, psat_soil )           # Water vapour density at soil (pwat_soil = psat_soil as we consider hum = 100% there)
    rhowat_air  = vapour_density( Tair, pwat_air )          # Water vapour density in the air

    qeva = d_qeva = 0.0
    if rhowat_soil.value >= rhowat_air.value :
        # There is heat transfer only if there is a positive gradient of water vapour density
        Kconv_air = air_thermal_conductivity( tmean )
        Kdiff_air = air_thermal_diffusivity( tmean, standard_Pair, Kconv_air )
        Kdiff_vap = vapour_diffusion_coefficient( tmean, standard_Pair )
        
        lewis = ValueError( Kdiff_air.value / Kdiff_vap.value, math.sqrt(square(Kdiff_air.error) + square(Kdiff_air.value * Kdiff_vap.error / Kdiff_vap.value)) / abs(Kdiff_vap.value) )
        rho_air = air_density( tmean, standard_Pair )
        cp_air  = air_spec_heat( tmean )

        h_frac = rho_air.value * cp_air.value * math.pow( lewis.value, 2./3. )
        hvap = hconv.value / h_frac
        d_hvap = math.sqrt( square(hconv.error) + square(hconv.value) * (square(rho_air.error/rho_air.value) + square(cp_air.error/cp_air.value) +
								         4. * square(lewis.error/lewis.value) / 9.) ) / h_frac

        qeva = wat_lh * hvap * ( rhowat_soil.value - rhowat_air.value )
        d_qeva = wat_lh * math.sqrt( square(d_hvap * (rhowat_soil.value - rhowat_air.value)) + square(hvap) * (square(rhowat_soil.error) + square(rhowat_air.error)) );
        
    return ValueError( qeva, d_qeva )

def calculateTSky4( Tbg, Tair, rHum, hconv_b ) :
    frac = this.ReEm * sb_constant

    heat_conv_b = convectiveHeat( Tbg, Tair, hconv_b )
    heat_evap_b = evaporativeHeat( Tbg, Tair, rHum, hconv_b )

    TSky4   = doublesquare( Tbg.value ) + (heat_conv_b.value + heat_evap_b.value) / frac
    d_TSky4 = math.sqrt( (4 * cube(Tbg.value) * square(Tbg.error)) + ((square(heat_conv_b.error) + square(heat_evap_b.error)) / square(frac)) )

    return ValueError( TSky4, d_TSky4 )

def radiativeHeat( Ts, TSky4 ) :
    frac = this.ReEm * sb_constant

    qrad   = frac * (doublesquare(Ts.value) - TSky4.value)
    d_qrad = frac * math.sqrt( (4 * cube(Ts.value) * square(Ts.error)) + square(TSky4.error) )

    return ValueError( qrad, d_qrad )

class RawPoint :
    def __init__( self, dataline=None ) :
        if dataline is None :
            self.pid = None
            self.date = self.strdate = None
            self.pt = None

            self.Tair = self.rHum = self.Ts = self.Tbg = self.Tg = None
            self.grad = None
            self.status = None

            self.good = False
        elif dataline[0] != '#' :
            parts = dataline.split()

            self.pid = parts[0]
            self.strdate = parts[1]
            self.date = datetime.datetime.strptime( self.strdate, "%Y/%m/%dT%H:%M:%S" )
            self.pt = [ float(parts[2]), float(parts[3]) ]

            self.Tair = ValueError( to_kelvin(float(parts[4])), 0.1 )
            self.rHum = ValueError( float(parts[5]), 1 )
            self.Ts   = ValueError( to_kelvin(float(parts[6])), float(parts[7]) )
            self.Tbg  = ValueError( to_kelvin(float(parts[8])), float(parts[9]) )
            self.Tg   = ValueError( to_kelvin(float(parts[10])), float(parts[11]) )
            self.grad = ValueError( 100. * float(parts[12]), 100. * float(parts[13]) )
            self.status = parts[14]

            self.good = True
        else :
            self.good = False

        return

class DataPoint :
    def __init__( self, raw=RawPoint() ) :
        if raw.good :
            self.pid = raw.pid
            self.date = raw.date
            self.strdate = raw.strdate
            self.pt = raw.pt

            self.Ts = raw.Ts
            self.Tair = raw.Tair
            if (self.Ts is not None) and (self.Tair is not None) :
                self.hconv = convective_heat_coefficient( self.Ts, self.Tair, this.lScale )
            else :
                self.hconv = None

            self.rHum = raw.rHum
            self.Tbg = raw.Tbg
            if (self.Tbg is not None) and (self.Tair is not None) and (self.rHum is not None) :

                if this.tairassky :
                    self.hconv_b = 42 # Not used anymore but it cannot be None
                    self.TSky4 = doublesquare( self.Tair )
                else :
                    self.hconv_b = convective_heat_coefficient( self.Tbg, self.Tair, this.lScale )
                    self.TSky4 = calculateTSky4( self.Tbg, self.Tair, self.rHum, self.hconv_b )
            else :
                self.hconv_b = None
                self.TSky4   = None

            self.Tg = raw.Tg
            self.grad = raw.grad
            self.status = raw.status

            self.Heat = None
        else :
            self.good = False
            
        return

    def _calculate_heat( self ) :
        self.Heat = [ self.convective_heat(), self.evaporative_heat(), self.radiative_heat() ]
        self.Heat += [ ValueError.sum_l2( self.Heat ) ]

        return

    @classmethod
    def header( cls ) :
        hdr = "{:<8}".format( "#Id" ) + \
            "{:2}".format( "F" ) + \
            "{:<21}".format( "Date" ) + \
            "{:9}".format( "X" ) + \
            "{:10}".format( "Y" ) + \
            "{:>8}".format( "Ts" ) + \
            "{:>9}".format( "d(Ts)" ) + \
            "{:>8}".format( "Tg" ) + \
            "{:>10}".format( "d(Tg)" ) + \
            "{:>9}".format( "dT/dz" ) + \
            "{:>9}".format( "d(dT/dz)" ) + \
            "{:>10}".format( "Qc" ) + \
            "{:>10}".format( "d(Qc)" ) + \
            "{:>10}".format( "Qe" ) + \
            "{:>10}".format( "d(Qe)" ) + \
            "{:>10}".format( "Qr" ) + \
            "{:>10}".format( "d(Qr)" ) + \
            "{:>10}".format( "Qtot" ) + \
            "{:>10}".format( "d(Qtot)" )

        return hdr

    def point_heat( self ) :
        if self.Heat is None :
            self._calculate_heat()

        return self.Heat

    def line( self ) :
        if self.Heat is None :
            self._calculate_heat()

        ( Qconv, Qeval, Qrad, Qtot ) = self.Heat

        if self.grad.value >= 1000 :
            grad_format = "{:>10.4f}"
        else :
            grad_format = "{:>9.4f}"

        line = "{:<8}".format( self.pid ) + \
            "{:<2}".format( self.status ) + \
            "{:<21}".format( self.strdate ) + \
            "{:<9.1f}".format( self.pt[0] ) + \
            "{:<10.1f}".format( self.pt[1] ) + \
            "{:>8.4f}".format( self.Ts.value ) + \
            "{:>9.4f}".format( self.Ts.error ) + \
            "{:>9.4f}".format( self.Tg.value ) + \
            "{:>9.4f}".format( self.Tg.error ) + \
            grad_format.format( self.grad.value ) + \
            "{:>9.4f}".format( self.grad.error ) + \
            "{:>10.4f}".format( Qconv.value ) + \
            "{:>10.4f}".format( Qconv.error ) + \
            "{:>10.4f}".format( Qevap.value ) + \
            "{:>10.4f}".format( Qevap.error ) + \
            "{:>10.4f}".format( Qrad.value ) + \
            "{:>10.4f}".format( Qrad.error ) + \
            "{:>11.4f}".format( Qtot.value ) + \
            "{:>9.4f}".format( Qtot.error )

        return line

    def good( self ) :
        return self.good and (self.hconv is not None) and (self.hconv_b is not None) and (self.TSky4 is not None) and (self.Tair.value > -274) and (self.rHum.value <= 100)

    def convective_heat( self ) :
        if (self.Ts is not None) and (self.Tair is not None) and (self.hconv is not None) :
            value = convectiveHeat( self.Ts, self.Tair, self.hconv )
        else :
            value = ValueError( float('NaN'), float('NaN') )

        return value

    def evaporative_heat( self ) :
        if (self.Ts is not None) and (self.Tair is not None) and (self.rHum is not None) :
            value = evaporativeHeat( self.Ts, self.Tair, self.rHum, self.hconv )
        else :
            value = ValueError( float('NaN'), float('NaN') )

        return value

    def radiative_heat( self ) :
        if (self.Ts is not None) and (self.TSky4 is not None) :
            value = radiativeHeat( self.Ts, self.TSky4 )
        else :
            value = ValueError( float('NaN'), float('NaN') )

        return value

class DataPixel :
    Tair = ValueError( TAMB + AZO, 0.1 )   # Air temperature
    rHum = ValueError( 10., 1 )            # Relative humidity
    Tbg  = ValueError( TAMB + AZO, 0.1 )   # Background temperature
    hconv_b     = None       # Convective heat factor of background
    TSky4       = None       # Sky temperature ^ 4
    
    def __init__( self, x, y, Ts, celsius=True ) :
        self.pt = [x, y]
        self.Ts = Ts
        if celsius :
            self.Ts.value = to_kelvin( Ts.value )

        self.hconv = convective_heat_coefficient( self.Ts, DataPixel.Tair, this.lScale )

        return

    @classmethod
    def set_environment( cls, Tbg, Tair, rHum ) :
        cls.Tair = Tair
        cls.Tbg  = Tbg
        cls.rHum = rHum
        
        if this.tairassky :
            cls.hconv_b = 42 # Not used anymore but it cannot be None
            cls.TSky4 = doublesquare( cls.Tair )
        else :
            cls.hconv_b = convective_heat_coefficient( cls.Tbg, cls.Tair, this.lScale )
            cls.TSky4 = calculateTSky4( cls.Tbg, cls.Tair, cls.rHum, cls.hconv_b )

        return

    def convective_heat( self ) :
        return convectiveHeat( self.Ts, DataPixel.Tair, self.hconv )

    def evaporative_heat( self ) :
        return evaporativeHeat( self.Ts, DataPixel.Tair, DataPixel.rHum, self.hconv )

    def radiative_heat( self ) :
        return radiativeHeat( self.Ts, DataPixel.TSky4 )

class ThermalCluster :
    def __init__( self, c, image, area ) :
        self.cluster = c
        self.cellarea = area

        self.flux = np.array( [ [ pt[0], pt[1], ValueError(image.flux[pt[0], pt[1]], image.dflux[pt[0], pt[1]]) ] for pt in c.points ] )
        self.power = area * ValueError.sum_l2( self.flux[:, 2] )

        return

    def area( self ) :
        return self.cellarea * len(self.flux)

    def avg_power( self ) :
        return self.power / (self.cellarea * len(self.flux))

    def to_pixel( self ) :
        return Pixel( self.cluster.bary[0], self.cluster.bary[1], self.power.value )

    def fill_image( self, image, mean=False, value=None ) :
        if mean :
            power = self.avg_power()
        elif value is not None :
            power = value
        else :
            power = self.power

        for pt in self.flux :
            image[pt[0], pt[1]] = power.value

        return self

class ThermalImage :
    def __init__( self, image, error=2, celsius=True ) :
        data = np.array( image )

        self.shape = data.shape
        self.data = data.reshape( (-1, 1) ).flatten()
        self.celsius = celsius
        self.error = error

        self.totpwr = self.redpwr = self.totnr = self.rednr = None
        self.redmask = None
        self.sigma = None

        return

    def __flux( self, t ) :
        if self.celsius :
            T = ValueError( to_kelvin(t), camera_error(t, self.error) )
        else :
            T = ValueError( t, camera_error(to_celsius(t), self.error) )

        if this.onlyrad :
            result = radiativeHeat( T, self.TSky4 )
        else :
            hconv = convective_heat_coefficient( T, self.Tair, this.lScale )

            Qconv = convectiveHeat( T, self.Tair, hconv )
            Qevap = evaporativeHeat( T, self.Tair, self.rHum, hconv )
            Qrad  = radiativeHeat( T, self.TSky4 )

            result = ValueError.sum_l2( [Qconv, Qevap, Qrad] )

        return result

    def update_image( self, image ) :
        data = np.array( image )

        self.shape = data.shape
        self.data = data.reshape( (-1, 1) ).flatten()

        return self

    def air_temp( self ) :
        temp = to_celsius(self.Tair.value) if self.celsius else self.Tair.value

        return ValueError( temp, self.Tair.error )

    def bg_temp( self ) :
        temp = to_celsius(self.Tbg.value) if self.celsius else self.Tbg.value

        return ValueError( temp, self.Tbg.error )

    def calculate_flux( self, Tbg, Tair, rHum, mask=None ) :
        if self.celsius :
            self.Tbg = ValueError( to_kelvin(Tbg.value), Tbg.error )
            self.Tair = ValueError( to_kelvin(Tair.value), Tair.error )
        else :
            self.Tbg = Tbg
            self.Tair = Tair
        self.rHum = rHum

        if this.tairassky or this.onlyrad :
            self.hconv_b = 42 # Not used anymore but it cannot be None
            self.TSky4 = doublesquare( self.Tair )
        else :
            self.hconv_b = convective_heat_coefficient( self.Tbg, self.Tair, this.lScale )
            self.TSky4 = calculateTSky4( self.Tbg, self.Tair, self.rHum, self.hconv_b )

        if self.celsius :
            tbg  = to_celsius( Tbg.value )
            tair = to_celsius( Tair.value )
        else :
            tbg  = Tbg.value
            tair = Tair.value

        if mask is None :
            if this.tairassky or this.onlyrad :
                couples = [ self.__flux( t ) if not np.isnan(t) and (t > tair) else ValueError(0, 0) for t in self.data ]
            else :
                couples = [ self.__flux( t ) if not np.isnan(t) and (t > tbg) else ValueError(0, 0) for t in self.data ]

            data = np.array( [ c.value for c in couples ] )
            errs = np.array( [ c.error for c in couples ] )

            self.minimum = data[data > 0].min()
            self.maximum = data.max()

            mask = np.isnan( self.data )
            data[mask] = float( 'NaN' )
            errs[mask] = float( 'NaN' )

            self.flux = data.reshape( self.shape )
            self.dflux = errs.reshape( self.shape )

        else :
            flxmask = mask.reshape( (-1, 1) ).flatten()

            couples = [ self.__flux(t) if y else ValueError(0, 0) for (y, t) in zip(flxmask, self.data) ]
            data = np.array( [ c.value for c in couples ] )
            errs = np.array( [ c.error for c in couples ] )

            flux = self.flux.reshape( (-1, 1) ).flatten()
            dflx = self.dflux.reshape( (-1, 1) ).flatten()

            flux[flxmask] = data[flxmask]
            dflx[flxmask] = errs[flxmask]

            self.flux  = flux.reshape( self.shape )
            self.dflux = dflx.reshape( self.shape )
            self.redmask = mask.reshape( self.shape )

            self.totpwr = self.redpwr = None

        return self

    def reduced_mask( self, area, sigma ) :
        if (self.redmask is None) or (self.sigma is None) or (self.sigma != sigma) :
            flux = self.flux.copy().reshape( (-1, 1) ).flatten()
            dflx = self.dflux.copy().reshape( (-1, 1) ).flatten()

            nanmask = np.isnan(flux)
            flux[nanmask] = 0.0
            dflx[nanmask] = 0.0

            zeromask = flux <= 0.0
            dzermask = dflx <= 0.0

            # Remove all point with negative flux
            # or zero flux error
            mask = zeromask | dzermask
            flux[mask] = 1.0
            dflx[mask] = sigma + 1

            ratio = np.divide( dflx, flux )
            sigmask = ratio < (1. / sigma)

            ratio[mask] = ratio.max()
            self.ratio = ratio

            self.redmask = sigmask.reshape( self.shape )
            self.sigma = sigma

        return self.redmask, self.ratio

    def _masked_power( self, mask ) :
        mask = mask.reshape( (-1, 1) ).flatten()
        flux = self.flux.reshape( (-1, 1) ).flatten()
        dflx = self.dflux.reshape( (-1, 1) ).flatten()
        
        flux = flux[mask]
        dflx = dflx[mask]
        
        array = [ ValueError(x[0], x[1]) for x in zip(flux, dflx) ]

        return ValueError.sum_l2( array )

    def total_power( self, area ) :
        if self.totpwr is None :
            flux = np.array( self.flux )
            flux[np.isnan(flux)] = 0
            pos_mask = flux > 0

            self.totpwr = self._masked_power( pos_mask )
            self.totnr = flux[pos_mask].size

        return area * self.totpwr

    def total_flux( self, area ) :
        self.total_power( area )

        return self.totpwr / self.totnr

    def reduced_power( self, area, sigma=None ) :
        if sigma is None :
            sigma = self.sigma

        if (self.redpwr is None) or (self.sigma is None) or (self.sigma != sigma) :
            mask, _ = self.reduced_mask( area, sigma )

            self.redpwr = self._masked_power( mask )
            self.rednr = self.flux[mask].size

        return area * self.redpwr

    def reduced_flux( self, area, sigma=None ) :
        self.reduced_power( area, sigma )

        return self.redpwr / self.rednr

    def cluster_flux( self, in_clus, area ) :
        out_clus = [ ThermalCluster(c, self, area) for c in in_clus ]

        return out_clus
