#!/usr/bin/python3
#  -*- coding: utf-8 -*-

import os
import sys
import re
import csv
import argparse
import datetime

import logger
import flight

def normalize_arguments( args ) :
    if args.noimages :
        args.imgdir = None

    args.outdir = args.output
    args.logfile = os.path.abspath( os.path.join(args.outdir, os.path.pardir, "multiflight.log") )

    return vars(args)

def do_flight( filename, arguments, prod, cmd=None ) :
    logger.clearpass()
    filepath = os.path.dirname( filename )

    if filepath == "" :
        if arguments['directory'] is None :
            indir = os.path.dirname( arguments['input'] )
            filepath = os.path.join( indir, filename )
        else :
            filepath = os.path.join( arguments['directory'], filename )

        if os.path.exists(filepath) :
            flight.single_flight( filepath, arguments, prod, cmd )
        else :
            logger.log( "Il file <{}> non esiste, saltato...".format(filepath) )
    elif os.path.exists(filename) :
        flight.single_flight( filename, arguments, prod )
    else :
        logger.log( "Il file <{}> non esiste, saltato...".format(filename) )

    return

def check_product_number( filename, pn, products, excluded ) :
    result = False
    
    if (products is not None) and (pn in products) :
        result = True
    elif (excluded is not None) and (pn in excluded) :
        print( "Il file <{}> è stato esplicitamente escluso dall'analisi in quanto corrisponde ad un prodotto non voluto ({})"
               .format(filename, ", ".join([str(p) for p in excluded])) )
    elif products is None :
        result = True
    else :
        print( "Il file <{}> non corrisponde ad un prodotto richiesto, saltato...".format(filename) )

    return result

def read_datfile( inputfile, products, excluded ) :
    product = re.compile( "^.+[Pp]rodotto_?([0-9]+)_.+$" )
    fileline = re.compile( "^\s*([^\s]+)\s+(-?[0-9\.eE]+)\s+(-?[0-9\.eE]+)\s*$" )
    comment = re.compile( "^\s*#.*$" )
    datalist = None

    with open( inputfile, 'r' ) as handle :
        datalist = []
        linen = 0
        
        for line in handle :
            linen += 1

            if comment.match(line) :
                continue

            result = fileline.match( line )
            if result :
                filename = result.group( 1 )
                tair = float( result.group(2) )
                hum  = float( result.group(3) )

                if (products is None) and (excluded is None) :
                    datalist.append( [ filename, tair, hum ] )
                else :
                    result = product.match( filename )

                    if result :
                        if check_product_number(filename, int(result.group(1)), products, excluded) :
                            datalist.append( [ filename, tair, hum ] )
                    else :
                        print( "Il file <{}> non definisce un prodotto, saltato...".format(filename) )
            else :
                print( "Error di sintassi nel file <{}> alla linea {}: \"{}\"".format(arguments['input'], linen, line) )
                exit( -1 )

    return datalist

def read_csvfile( inputfile, products, excluded, mh=None ) :
    hour = re.compile( "^([0-9]+):([0-9]+)$" )
    date = re.compile( "^([0-9]{4})([0-9]{2})([0-9]{2})$" )
    pisciarelli = re.compile( "^.*Pisciarelli.*$" )
    datalist = None

    with open( inputfile, 'r', newline='' ) as handle :
        datalist = []
        reader = csv.reader( handle, delimiter=';' )

        for row in reader :
            if row[0].isnumeric() :
                pn = int( row[0] )
                name, ext = os.path.splitext( row[1] )

                if not pisciarelli.match(name) :
                    print( "Il prodotto {} ({}) non è di Pisciarelli".format(pn, row[1]) )
                    continue

                if ext == ".txt" :
                    hr = hour.match( row[3] )
                    dr = date.match( row[4] )

                    if hr and dr :
                        filetime = datetime.datetime( int(dr.group(1)), int(dr.group(2)), int(dr.group(3)), int(hr.group(1)), int(hr.group(2)) )

                        if (filetime.hour > 8) and (filetime.hour < 16) :
                            print( "Il prodotto {} ({}) è stato acquisito di giorno ({}). Saltato...".format(pn, row[1], filetime) )
                            continue

                        try :
                            tair = float( row[5].replace(",", ".") )
                            rhum = float( row[6].replace(",", ".") )
                            height = float( row[7].replace(",", ".") )

                            if ((mh is not None) or (mh > 0)) and (height > mh) :
                                print( "Prodotto {} ({}) è stato acquisito ad un'altezza di {}m, superiore alla maggiore richiesta ({}m), saltato".format(pn, row[1], height, mh) )
                                continue

                            if ((products is None) and (excluded is None)) or check_product_number(row[1], pn, products, excluded) :
                                datalist.append( [ row[1], tair, rhum, filetime, height, pn ] )
                                print( "Aggiunto prodotto {} ({}), data: {}, T: {}, H: {}, Q: {}".format(pn, row[1], filetime, tair, rhum, height) )

                        except ValueError as e:
                            print( "Errore di sintassi in Tair o rHum del prodotto {}: <\"{}\" - \"{}\">".format(pn, row[5], row[6]) )
                            exit( -1 )
                    else :
                        print( "Errore di sintassi nella data e nell'ora del prodotto {}: <\"{}\" - \"{}\">".format(pn, row[3], row[4]) )
                        exit( -1 )
                else :
                    print( "Il file <{}> non è un ASCII GRID.".format(row[1]) )

        datalist.sort( key=lambda x: x[3] )

    return datalist

def execute( arguments, cmd=None ) :
    datfile = re.compile( "^.+\.dat$" )
    csvfile = re.compile( "^.+\.csv$" )

    inputfile = arguments['input']
    datalist = None

    if datfile.match(inputfile) :
        datalist = read_datfile( inputfile, arguments['products'], arguments['excluded'] )

        if datalist is None :
            print( "Errore durante la lettura del file DAT <{}>".format(inputfile) )
    elif csvfile.match(inputfile) :
        datalist = read_csvfile( inputfile, arguments['products'], arguments['excluded'], arguments['maxheight'] )

        if datalist is None :
            print( "Errore durante la lettura del file CSV <{}>".format(inputfile) )
    else :
        print( "File <{}> does not seem to be in a known format".format(inputfile) )

    if not arguments["onlyparse"] and (datalist is not None) :
        logger.openfile( arguments["logfile"], True )
        
        logger.log( "File di descrizione <{}> letto con successo, processo i dati...".format(inputfile) )
        logger.log( "<--------------------------------------------------------------->" )

        for l in datalist :
            arguments['tair']   = l[1]
            arguments['hum']    = l[2]
            arguments['height'] = l[4]

            do_flight( l[0], arguments, l[5], cmd )

    return

parser = argparse.ArgumentParser( description="Il programma fornisce la stima del flusso di calore di uno o più file in formato ASCII GRID" )

parser.add_argument( '-i', "--input", help="Path del file contenente l'elenco dei file da esaminare",
                     metavar="FILENAME", type=str, required=True )
parser.add_argument( '-O', "--output", help="Path della directory dove devono essere scritti i file di output",
                     metavar="OUTPUT", type=str, default="./output" )
parser.add_argument( '-I', "--imgdir", help="Path della directory dove devono essere scritte le immagini di output",
                     metavar="IMGDIR", type=str, default="./images" )
parser.add_argument( '-s', "--segments", help="Numero di segmenti in cui dividere l'immagine per l'analisi preliminare",
                     metavar="SEGMENTS", type=int, default=3 )
parser.add_argument( '-r', "--radius", help="Massima distanza tra i pixel nello spazio 3D per essere considerati apparteneneti al medesimo cluster",
                     metavar="RADIUS", type=float, default=10 )
parser.add_argument( '-m', "--minsize", help="Minimo numero di pixel in un cluster",
                     metavar="MINSIZE", type=int, default=4 )
parser.add_argument( '-n', "--noisedist", help="Massima distanza tra un pixel ed un cluster per la successiva riaggregazione",
                     metavar="NOISEDIST", type=float, default=100 )
parser.add_argument( '-d', "--directory",
                     type=str, help="Directory in cui trovare i file di input. Se non settata corrisponde utilizza la directory in cui si trova FILENAME" )
parser.add_argument( '-e', "--vuerror", help="Errore percentuale sui pixel della termocamera",
                     metavar="VUERROR", type=float, default=5. )
parser.add_argument( '-S', "--sigma", help="Distanza sigma per considerare un punto di flusso diverso da zero",
                     metavar="SIGMA", type=float, default=2. )
parser.add_argument( '-L', "--loss", help="Perdita percentuale di temperatura per metro di altezza. Se non impostato non viene corretto",
                     metavar="LOSS", type=float, default=0. )
parser.add_argument( '-c', "--mkdir", action="store_true", help="Crea le directory a partire dalle date in IMGDIR e OUTPUT per ogni prodotto" )
parser.add_argument( '-D', "--draw", action="store_true", help="Quando impostato le immagini vengono anche mostrate a schermo" )
parser.add_argument( '-N', "--noimages", action="store_true", help="Quando impostato i file di immagini non vengono creati" )
parser.add_argument( '-W', "--warp", action="store_true", help="Applica il \"warping\" dei cluster prima della generazione dei supercluster. (Al momento non usato)" )
parser.add_argument( '-V', "--verbose", action="store_true", help="Output verboso: vengono creati solo i file di log" )
parser.add_argument( '-P', "--products", nargs='+', help="Analizza solamente i prodotti il cui numero è elencato di seguito", metavar="PRODID", type=int )
parser.add_argument( '-E', "--excluded", nargs='+', help="Esclude dall'analisi i prodotti riportati di seguito", metavar="PRODID", type=int )
parser.add_argument( '-B', "--bgpoint", nargs=3, help="Utilizza il punto ed il raggio passati come i tre parametri per calcoare la temperatura del background", metavar="BGPOINT", type=float, default=None )
parser.add_argument( '-k', "--skyair", action="store_true", help="Utilizza la temperatura dell'aria al posto della Tsky. Non calcola il background" )
parser.add_argument( '-R', "--radiative", action="store_true", help="Usa solamente la parte radiativa del flusso. Impone l'utilizzo della temperatura dell'aria come Tsky. Non calcola il background" )
parser.add_argument( "--nogrow", action="store_true", help="I pixel di rumore non vengono recuperati durante la creazione dei supercluster." )
parser.add_argument( "--norecluster", action="store_true", help="Non viene effettuato il secondo reclustering" )
parser.add_argument( "--borders", help="I bordi dei supercluster sono rimossi dall'immagine. BSIZE è la dimensione del bordo che si vuole rimuovere", metavar="BSIZE", type=int, default=0 )
parser.add_argument( "--defaults", action="store_true", help="Stampa i parametri di default ed esce." )
parser.add_argument( "--maxheight", help="Discards flights over HEIGHT meters. All files are included if not set", metavar="HEIGHT", type=float, default=0 )
parser.add_argument( "--onlyparse", help="Only reads the input FILENAME and exits", action="store_true" )
parser.add_argument( "--large", help="Create images with biggest fonts for definitive article output", action="store_true" )

arguments = parser.parse_args()
arguments = normalize_arguments( arguments )

if arguments['defaults'] :
    print( "Default argument values:" )
    
    for arg in arguments :
        print( "\t{}: {}".format(arg, arguments[arg]) )
else :
    execute( arguments, " ".join(sys.argv) )
