#!/usr/bin/python3
#  -*- coding: utf-8 -*-

import sys
import re
import os
import csv

import datetime
import pyproj
import numpy as np

import logger
import asciigrid
from utils import ValueError

this = sys.modules[__name__]
this.verbose = False

def set_verbose( v=True ) :
    this.verbose = v

    return

class Output :
    good = False
    start = end = outdir = noext = date = None
    data = None
    product = "--"
    
    def __init__( self, outdir, filename, date ) :
        filename = os.path.basename( filename )
        
        expr = re.compile( "^(.+)_[tT]_(.+)$" )
        result = expr.match( filename )

        if result :
            self.good = True

            self.start = result.group( 1 )
            self.end   = result.group( 2 )

            self.filename = filename
            self.outdir = outdir
            self.date = date

            self.gridfiles = []

        fname = re.compile( "^(.+(?:(?<!\.txt)))(?:\.txt)+$" )
        result = fname.match( self.end )

        if result :
            self.noext = result.group( 1 )
        else :
            self.good = False

        return

    def _create_cluster_mask( self, shape ) :
        image = np.array( [ float('NaN') ] * (shape[0] * shape[1]) ).reshape( shape )
        clusters = self.data['clus'].get_flux_clusters()

        for i in range(len(clusters)) :
            c = clusters[i]
            
            for pt in c.cluster.points :
                image[pt[0], pt[1]] = float( i )

        return image

    def _create_cluster_rows( self ) :
        #proj = pyproj.Proj( "+proj=utm +zone=33t, +north +ellps=WGS84 +datum=WGS84 +units=m +no_defs" )
        proj = pyproj.Proj( proj="utm", zone=33, ellpl="WGS84", datum="WGS84", units="m" )
        grid = self.data['grid']
        clusters = self.data['clus'].get_flux_clusters()

        lines = [ [ "CID", "UTM_x", "UTM_y", "Lon", "Lat" ] ]

        for i in range(len(clusters)) :
            c = clusters[i]

            px = grid.pixel2utm( c.to_pixel(), True )
            lines.append( [ i, px.x(), px.y(), *proj(px.x(), px.y(), inverse=True) ] )

        return lines

    def _prepare_dat_header( self ) :
        grid = self.data['grid']
        flux = self.data['thermal']

        total = flux.total_power( grid.cellarea() )
        redux = flux.reduced_power( grid.cellarea() )

        totfl = flux.total_flux( grid.cellarea() )
        redfl = flux.reduced_flux( grid.cellarea() )
        
        header = [ ( "# H{:02d} Elaboration data - filename date product",
                     "H{{:02d}} \"{}\" {} {:d}".format(self.filename, self.date.strftime("%Y/%m/%d"), self.product) )
        ]

        if len(self.gridfiles) != 0 :
            for cp in self.gridfiles :
                what = cp[0].replace( ' ', '_' )
                header.append( ("# H{{:02d}} grid file - {}".format(what), "H{{:02d}} \"{}\"".format(cp[1])) )

        tmax = grid.maximum
        if tmax > 100 :
            dtmax = flux.error * tmax / 100.
        else :
            dtmax = flux.error

        header.extend( [
            ( "# H{:02d} Image data - columns rows width height cellsize area used_pixels",
              "H{{:02d}} {:d} {:d} {:f} {:f} {:f} {:f} {}".format(grid.columns, grid.rows, grid.width(), grid.height(), grid.cellsize, grid.size(), grid.countgood) ),
            ( "# H{:02d} Env data - camera_error Tair dTair rHum drHum",
              "H{{:02d}} {:f} {:f} {:f} {:f} {:f}".format(flux.error, flux.air_temp().value, flux.air_temp().error, flux.rHum.value, flux.rHum.error) ),
            ( "# H{:02d} Global results - Tmax dTmax",
              "H{{:02d}} {:f} {:f}".format(tmax, dtmax) ),
            ( "# H{:02d} Global results - Tmean dTmean Tbg dTbg Ptot dPtot Pred dPred Ftot dFtot Fred dFred",
              "H{{:02d}} {:f} {:f} {:f} {:f} {:f} {:f} {:f} {:f} {:f} {:f} {:f} {:f}".format(grid.average().value, grid.average().error, flux.bg_temp().value, flux.bg_temp().error,
                                                                                             total.value, total.error, redux.value, redux.error,
                                                                                             totfl.value, totfl.error, redfl.value, redfl.error) )
        ] )

        return header

    def _prepare_dat_data( self ) :
        #proj = pyproj.Proj( "+proj=utm +zone=33t, +north +ellps=WGS84 +datum=WGS84 +units=m +no_defs" )
        proj = pyproj.Proj( proj="utm", zone=33, ellpl="WGS84", datum="WGS84", units="m" )
        grid = self.data['grid']
        clusters = self.data['clus'].get_flux_clusters()
        number = len( clusters )
        
        lines = [ "#", "# D00 hot clusters - ID name barx bary UTM_x UTM_y Lon Lat Rp dRp Rm dRm Ap Am2 Tm dTm P dP F dF" ]

        for i in range(number) :
            c = clusters[i]
            tc = c.cluster
            px = grid.pixel2utm( c.to_pixel(), True )

            avg = c.avg_power()

            data = [ "C{:02d}".format(i), "{:03d}:{:03d}".format(tc.segment, tc.cluster),
                     tc.bary[0], tc.bary[1],
                     px.x(), px.y(),
                     *proj(px.x(), px.y(), inverse=True),
                     tc.rmean, tc.drms,
                     grid.cellsize * tc.rmean, grid.cellsize * tc.drms,
                     tc.size(), c.area(),
                     tc.tmean, tc.trms,
                     c.power.value, c.power.error,
                     avg.value, avg.error ]

            line = ' '.join( [str(d) for d in data] )
            lines.append( line )
        
        return lines      

    def _prepare_csv_summary( self ) :
        tform = "%Y/%m/%d %H:%M:%S %Z"
        now = datetime.datetime.now( datetime.timezone.utc )
        #proj = pyproj.Proj( "+proj=utm +zone=33T, +north +ellps=WGS84 +datum=WGS84 +units=m +no_defs" )
        proj = pyproj.Proj( proj="utm", zone=33, ellpl="WGS84", datum="WGS84", units="m" )
        
        grid = self.data['grid']
        flux = self.data['thermal']

        total = flux.total_power( grid.cellarea() )
        redux = flux.reduced_power( grid.cellarea() )

        totfl = flux.total_flux( grid.cellarea() )
        redfl = flux.reduced_flux( grid.cellarea() )

        clus = self.data['clus']
        clusters = clus.get_flux_clusters()

        tmax = grid.maximum
        if tmax >= 100 :
            dtmax = flux.error * tmax / 100.
        else :
            dtmax = flux.error
        Tmax = ValueError( tmax, dtmax )
        
        lines = [
            [ "Fine elaborazione", now.strftime(tform), now.astimezone().strftime(tform), "Prodotto", self.product, "Data volo", self.date.strftime("%Y/%m/%d"), "File di input", self.filename ],
            [ "" ],
            [ "", "Larghezza (px)", "Altezza (px)", "Larghezza (m)", "Altezza (m)", "m/pixel", "Area totale (pixel)", "Area utile (pixel)", "Area utile (m²)" ],
            [ "Parametri immagine", grid.columns, grid.rows, grid.width(), grid.height(), grid.cellsize, grid.size(), grid.countgood, grid.used_surface() ],
            [ "Errore camera (± °C/%)", flux.error ],
            [ "", "Valore (°C)", "Errore (°C)" ],
            [ "Temperatura media", grid.average().value, grid.average().error, grid.average() ],
            [ "Temperatura massima", tmax, dtmax, Tmax ],
            [ "" ],
            [ "", "Tair (°C)", "dTair (°C)", "rHum (%)", "drHum (%)" ],
            [ "Parametri ambientali", flux.air_temp().value, flux.air_temp().error, flux.rHum.value, flux.rHum.error ],
            [ "" ],
            [ "Sommario risultati" ],
            [ "", "Valore", "Errore", "Unità", "", "Sigma cut" ],
            [ "Temperatura background", flux.bg_temp().value, flux.bg_temp().error, "°C", "{} °C".format(flux.bg_temp()) ],
            [ "Potenza totale emessa", total.value, total.error, "W", "{} W".format(total) ],
            [ "Potenza totale ridotta", redux.value, redux.error, "W", "{} W".format(redux), flux.sigma ],
            [ "Flusso medio totale", totfl.value, totfl.error, "W/m²", "{} W/m²".format(totfl) ],
            [ "Flusso medio ridotto", redfl.value, redfl.error, "W/m²", "{} W/m²".format(redfl), flux.sigma ],
            [ "" ],
            [ "Cluster caldi identificati", len(clusters) ],
            [ "", "", "Baricentro" ] + [ "" ]*9 + [ "Superficie", "", "Temperatura", "", "Rilascio" ],
            [ "ID", "Seg:Ncl",
              "Px", "Py",
              "UTM_x", "UTM_y",
              "Lon", "Lat",
              "Rm (px)", "dRm (px)",
              "Rm (m)", "dRm (m)",
              "Area (px)", "Area (m²)",
              "Tm (°C)", "dTm (°C)",
              "P (W)", "dP (W)",
              "F (W/m²)", "dF (W/m²)"
             ]
        ]

        for i in range(len(clusters)) :
            c = clusters[i]
            tc = c.cluster
            px = grid.pixel2utm( c.to_pixel(), True )

            avg = c.avg_power()

            (lon, lat) = proj( px.x(), px.y(), inverse=True )
            lon = "{:.7f}".format( lon )
            lat = "{:.7f}".format( lat )

            li = [ i, "{:03d}:{:03d}".format(tc.segment, tc.cluster),
                   tc.bary[0], tc.bary[1],
                   px.x(), px.y(),
                   lon, lat,
                   tc.rmean, tc.drms,
                   grid.cellsize * tc.rmean, grid.cellsize * tc.drms,
                   tc.size(), c.area(),
                   tc.tmean, tc.trms,
                   c.power.value, c.power.error,
                   avg.value, avg.error ]

            lines.append( li )

        return lines

    def set_product( self, prod ) :
        self.product = int(prod)

        return self

    def set_data( self, grid, clus, thermal ) :
        self.data = { 'grid': grid, 'clus': clus, 'thermal': thermal }

        return self

    def write_image( self, image, part, what ) :
        filepath = os.path.join( self.outdir, "{}_{}_{}".format(self.start, part, self.end) )
        
        if this.verbose :
            logger.log( "Writing {} to file <{}>...".format(what, filepath), end=' ', flush=True )

        self.writer.write_image( image, filepath )

        if this.verbose :
            logger.log( "Done!" )

        self.gridfiles.append( (what, filepath) )

        return

    def write_csv( self, lines, part, what, delm=';' ) :
        filepath = os.path.join( self.outdir, "{}_{}_{}.csv".format(self.start, part, self.noext) )

        if this.verbose :
            logger.log( "Writing {} to CSV file <{}>...".format(what, filepath), end=' ', flush=True )

        with open( filepath, "w" ) as handle :
            writer = csv.writer( handle, delimiter=delm )

            for l in lines :
                if delm == ';' :
                    writer.writerow( [ "{:.2f}".format(el).replace('.', ',') if isinstance(el, float) else el for el in l ] )
                else :
                    writer.writerow( l )

        if this.verbose :
            logger.log( "Done!" )

        return

    def write_dat( self, header, data ) :
        filepath = os.path.join( self.outdir, "{}_SUM_{}.dat".format(self.start, self.noext) )

        if this.verbose :
            logger.log( "Writing summary data to file <{}>...".format(filepath), end=' ', flush=True )

        with open( filepath, "w" ) as handle :
            for i in range(len(header)) :
                line = header[i][0].format( i )
                print( line, file=handle )

                line = header[i][1].format( i )
                print( line, file=handle )

            for line in data :
                print( line, file=handle )

        if this.verbose :
            logger.log( "Done!" )

        return

    def write_gridfiles( self ) :
        if this.verbose :
            logger.log( "Writing output grid files..." )

        shape = (self.data['grid'].rows, self.data['grid'].columns)
        ndmask = self.data['grid'].ndmask
        nodata = self.data['grid'].nodata
        
        self.writer = asciigrid.GridWriter( self.data['grid'] )

        image = np.array( self.data['thermal'].flux ).reshape( shape )
        image[ndmask] = nodata
        self.write_image( image, "HF", "heat flux map" )

        image = np.array( self.data['thermal'].dflux ).reshape( shape )
        image[ndmask] = nodata
        self.write_image( image, "dHF", "error on heat flux map" )

        image = self._create_cluster_mask( shape )
        image[ndmask] = nodata
        image[np.isnan(image)] = nodata
        self.write_image( image, "CM", "cluster mask" )

        return self

    def write_csvfiles( self ) :
        if this.verbose :
            logger.log( "Writing output CSV files..." )

        lines = self._create_cluster_rows()
        self.write_csv( lines, "CP", "clusters position", ',' )

        lines = self._prepare_csv_summary()
        self.write_csv( lines, "SUM", "results summary" )

        return self

    def write_datfiles( self ) :
        if this.verbose :
            logger.log( "Writing output DAT file..." )

        header = self._prepare_dat_header()
        data   = self._prepare_dat_data()

        self.write_dat( header, data )

        return self
