#!/usr/bin/python3
#  -*- coding: utf-8 -*-

import math
import pprint

import numpy as np

class Pixel :
    def __init__( self, col, row, val = None ) :
        self.pixel = ( col, row )
        self.value = val

        return

    def x( self ) :
        return self.pixel[0]

    def y( self ) :
        return self.pixel[1]

    def __str__( self ) :
        return "({}, {})".format(self.pixel[0], self.pixel[1]) if self.value is None else "({}, {}) => {:.1f}".format(self.pixel[0], self.pixel[1], self.value)

class ValueError :
    def __init__( self, value, error, weight=None ) :
        self.value = value
        self.error = error

        if weight is not None :
            self.weight = weight
        elif error != 0 :
            self.weight = 1 / (error * error)
        else :
            self.weight = 1

        return

    def __repr__( self ) :
        return pprint.pformat( vars(self) )

    def __add__( self, other ) :
        if isinstance(other, ValueError) :
            res = ValueError( self.value + other.value, self.error + other.error, self.weight )
        elif isinstance(other, float) or isinstance(other, int) :
            res = ValueError( self.value + other, self.error, self.weight )
        else :
            res = self

        return res

    def __sub__( self, other ) :
        if isinstance(other, ValueError) :
            res = ValueError( self.value - other.value, self.error + other.error, self.weight )
        elif isinstance(other, float) or isinstance(other, int) :
            res = ValueError( self.value - other, self.error, self.weight )
        else :
            res = self

        return res

    def __mul__( self, val ) :
        if self.weight is not None :
            other = ValueError( val * self.value, val * self.error, val * self.weight )
        else :
            other = ValueError( val * self.value, val * self.error )

        return other

    def __rmul__( self, val ) :
        return self.__mul__( val )

    def __truediv__( self, val ) :
        if self.weight is not None :
            other = ValueError( self.value / val, self.error / val, self.weight / val )
        else :
            other = ValueError( self.value / val, self.error / val )

        return other

    def __str__( self ) :
        value = self.value
        error = self.error
        pm = "±"

        if error == 0 :
            return "{:d} {} 0".format(int(value), pm)
        
        if value != 0 :
            exp_value = math.ceil( math.log10( abs(value) ) )
        else :
            exp_value = 1

        exp_error = math.floor( math.log10( error ) )
        number = math.trunc( error * math.pow(10, -exp_error) )

        if number == 1 :
            exp_error -= 1
            number = math.trunc( error * math.pow(10, -exp_error) )

        if exp_error < 0 :
            error_width = 2 - exp_error
            error_format = "{{:{}.{}f}}".format( error_width, -exp_error )

            if exp_value < 0 :
                value_width = 2 - exp_error
                value_format = "{{:{}.{}f}}".format( value_width, -exp_error )
            else :
                value_width = exp_value + 1 - exp_error
                value_format = "{{:{}.{}f}}".format( value_width, -exp_error )

            couple_format = "{} {} {}".format(value_format, pm, error_format)
            output = couple_format.format( value, error )
        else :
            error = int( math.trunc(number * math.pow(10, exp_error)) )
            value = int( math.trunc(math.pow(10, exp_error) * round(value / math.pow(10, exp_error))) )

            output = "{:d} {} {:d}".format( value, pm, error )

        return output

    def copy( self ) :
        return ValueError( self.value, self.error, self.weight )

    @classmethod
    def average( cls, array ) :
        data = np.array( array ) #.reshape( -1, 1 )

        if data.size != 0 :
            if isinstance(array[0], float) or isinstance(array[0], int) or (data.dtype == "int64") or (data.dtype == "int32") or (data.dtype == "float32") or (data.dtype == "float64") :
                m = data.sum() / data.size

                diff2 = np.array( [ (x - m) * (x - m) for x in data ] )
                s = math.sqrt( diff2.sum() / (data.size - 1) )
            else :
                values = np.array( [ d.value for d in data ] )
                weights = np.array( [ d.weight for d in data ] )
                wsum = weights.sum()

                m = np.dot( values, weights ) / wsum
                s = np.sqrt( 1 / wsum )
        else :
            m = s = 0

        return ValueError( m, s )

    @classmethod
    def sum_l1( cls, array ) :
        couples = np.array( [ [ve.value, ve.error] for ve in array ] )

        values = np.array( couples[:, 0] )
        errors = np.array( couples[:, 1] )

        return ValueError( values.sum(), errors.sum() )

    @classmethod
    def sum_l2( cls, array ) :
        values = np.array( [ ve.value for ve in array ] )
        errors = np.array( [ ve.error for ve in array ] )

        return ValueError( values.sum(), math.sqrt(np.dot(errors, errors)) )

    @classmethod
    def weight_mean( cls, array ) :
        hasw = True
        for ve in array :
            if ve.weight is None :
                hasw = False
                break

        couples = np.array( [ [ve.value, ve.error ] for ve in array ] )

        values = couples[:, 0]
        errors = couples[:, 1]

        if hasw :
            weights = np.array( [ ve.weight for ve in array ] )
            esum = np.dot( errors, weights )
        else :
            weights = np.reciprocal( np.square(errors) )

        msum = np.dot( values, weights )
        wsum = weights.sum()

        mean = msum / wsum

        if hasw :
            sigm = esum / wsum
        else :
            sigm = math.sqrt( 1 / wsum )

        return ValueError( mean, sigm )
